# coding: utf-8

"""
Unit test for monoexonic transcripts.
"""

import operator
import re
import unittest
from .. import exceptions, loci, parsers, transcripts
from ..loci import Transcript
from ..utilities.log_utils import create_null_logger, create_default_logger


class TranscriptTester(unittest.TestCase):
    tr_gff = """Chr1    TAIR10    mRNA    5928    8737    .    .    .    ID=AT1G01020.1;Parent=AT1G01020;Name=AT1G01020.1;Index=1
Chr1    TAIR10    exon    5928    8737    .    .    .    Parent=AT1G01020.1"""

    tr_lines = tr_gff.split("\n")
    for pos, line in enumerate(tr_lines):
        tr_lines[pos] = re.sub(r"\s+", "\t", line)
        assert len(tr_lines[pos].split("\t")) == 9, line.split("\t")

    tr_gff_lines = [parsers.GFF.GffLine(line) for line in tr_lines]

    for l in tr_gff_lines:
        assert l.header is False

    logger = create_null_logger("null")

    def setUp(self):
        """Basic creation test."""

        self.tr = Transcript()
        self.tr.logger = self.logger
        self.tr.chrom = "Chr1"
        self.tr.source = "TAIR10"
        self.tr.feature = "mRNA"
        self.tr.start = 5928
        self.tr.end = 8737
        self.tr.strand = "+"
        self.tr.add_exon((5928, 8737))
        self.tr.score = None
        self.tr.id, self.tr.parent, self.tr.name = "AT1G01020.1", "AT1G01020", "AT1G01020.1"
        self.tr.add_exon((8571, 8666), "CDS")
        self.tr.finalize()
        self.tr.configuration.pick.chimera_split.blast_check = False

        self.orf = parsers.bed12.BED12()
        self.orf.chrom = self.tr.id
        self.orf.start = 1
        self.orf.end = self.tr.cdna_length
        self.orf.name = self.tr.id
        self.orf.strand = "+"
        self.orf.score = 0
        self.orf.thick_start = 8571 - 5928 + 1
        self.orf.thick_end = 8666 - 5928 + 1
        self.orf.block_count = 1
        self.orf.blockSize = self.tr.cdna_length
        self.orf.block_starts = [0]
        self.orf.has_start_codon = True
        self.orf.has_stop_codon = True
        self.orf.transcriptomic = True
        self.assertFalse(self.orf.invalid, self.orf.invalid_reason)
        self.assertEqual((self.orf.thick_end - self.orf.thick_start + 1) % 3, 0)

    def test_basics(self):

        self.assertEqual(self.tr.chrom, "Chr1")
        self.assertEqual(self.tr.exon_num, 1)
        self.assertEqual(self.tr.monoexonic, True)
        self.assertEqual(self.tr.exon_num, len(self.tr.exons))
        self.assertEqual(self.tr.start, 5928)
        self.assertEqual(self.tr.end, 8737)
        self.assertEqual(self.tr.exons,
                         [tuple([5928, 8737])],
                         self.tr.exons)

    def test_cds(self):
        """Test the CDS features.
        Note that in a single-exon transcript with no strand, start_codon and stop_codon are defined as False.
        """

        self.tr.load_orfs([self.orf])
        self.assertEqual(self.tr.combined_cds, self.tr.selected_cds)

        self.assertEqual(self.tr.combined_cds,
                         [tuple([8571, 8666])],
                         self.tr.combined_cds)
        self.assertEqual(self.tr.selected_cds_start, 8571)
        self.assertEqual(self.tr.selected_cds_end, 8666)
        self.assertEqual(self.tr.has_start_codon, True)
        self.assertEqual(self.tr.has_stop_codon, True)

    def test_equality(self):

        new_transcript = self.tr.deepcopy()

        self.assertTrue(new_transcript == self.tr)

        new_transcript.strand = None
        self.assertFalse(new_transcript == self.tr)  # They have now a different strand

        new_transcript.unfinalize()
        new_transcript.strand = "+"  # It becomes a multiexonic transcript, so it must have a strand
        new_transcript.end = 9737

        new_exon = parsers.GFF.GffLine(self.tr_lines[-1])
        new_exon.strand = "+"
        new_exon.start = 9000
        new_exon.end = 9737
        new_transcript.add_exon(new_exon)

        new_transcript.finalize()
        self.assertTrue(new_transcript != self.tr)

    def test_mono_finalising(self):

        transcript_line = [line for line in self.tr_gff_lines if line.feature == "mRNA" ]
        self.assertEqual(len(transcript_line), 1,
                         "\n".join([str(line) for line in self.tr_gff_lines]))

        tr = loci.Transcript(transcript_line[0])
        exon_lines = [line for line in self.tr_gff_lines if
                      line.is_exon is True and "UTR" not in line.feature.upper()]
        tr.add_exons(exon_lines)
        tr.add_exon((8571, 8666), "CDS")

        tr.finalize()
        self.assertGreater(tr.three_utr_length, 0)
        self.assertGreater(tr.five_utr_length, 0)

    def test_invalid_transcript(self):
        lines = """Chr1\tTAIR10\tmRNA\t5928\t8737\t.\t.\t.\tID=AT1G01020.1;Parent=AT1G01020;Name=AT1G01020.1;Index=1
Chr1\tTAIR10\tCDS\t8571\t7500\t.\t.\t0\tParent=AT1G01020.1;
Chr1\tTAIR10\tCDS\t7503\t8666\t.\t.\t0\tParent=AT1G01020.1;
Chr1\tTAIR10\texon\t5928\t8737\t.\t.\t.\tParent=AT1G01020.1"""

        gff_lines = [parsers.GFF.GffLine(line) for line in lines.split("\n")]
        self.assertIsInstance(gff_lines[0], parsers.GFF.GffLine)
        checker = False
        if gff_lines[0].feature.endswith("transcript") or "RNA" in gff_lines[0].feature.upper():
            checker = True
        self.assertTrue(checker)
        self.assertTrue(gff_lines[0].is_transcript)
        transcript = loci.Transcript(gff_lines[0])

        transcript.logger = self.logger
        transcript.add_exons(gff_lines[1:])

        with self.assertRaises(exceptions.InvalidCDS):
            transcripts.transcript_methods.finalizing._check_cdna_vs_utr(transcript)

    def test_utr(self):

        self.assertEqual(self.tr.selected_internal_orf,
                         [("UTR", tuple([5928, 8570])),
                          ("exon", tuple([5928, 8737])),
                          ("CDS", tuple([8571, 8666]), 0),
                          ("UTR", tuple([8667, 8737]))],
                         "Right: {0}\nFound{1}".format([("UTR", 5928, 8570), ("CDS", 8571, 8666), ("UTR", 8667, 8737)],
                                                       self.tr.selected_internal_orf))
        self.assertEqual(self.tr.combined_utr, [tuple([5928, 8570]),
                                                tuple([8667, 8737])])
        self.assertEqual(self.tr.five_utr, [tuple([5928, 8570])],
                         self.tr.five_utr)
        self.assertEqual(self.tr.three_utr, [tuple([8667, 8737])])

    def test_utr_metrics(self):

        """Test for UTR exon num, start distance, etc."""

        self.assertEqual(self.tr.five_utr_num, 1)
        self.assertEqual(self.tr.three_utr_num, 1)
        self.assertEqual(self.tr.five_utr_length, 8570 + 1 - 5928)
        self.assertEqual(self.tr.three_utr_length, 8737 + 1 - 8667)
        self.assertEqual(self.tr.selected_start_distance_from_tss,
                         8571 - 5928,
                         self.tr.selected_end_distance_from_tes)
        self.assertEqual(self.tr.selected_end_distance_from_tes,
                         8737 - 8666,
                         (self.tr.selected_end_distance_from_tes, self.tr.strand))

    def test_strip_cds(self):

        self.tr.strip_cds()
        self.assertEqual(self.tr.selected_cds_length, 0)
        self.assertEqual(self.tr.three_utr, [])
        self.assertEqual(self.tr.five_utr, [])
        self.assertEqual(self.tr.selected_cds, [])
        self.assertEqual(self.tr.selected_cds_start, None)
        self.assertEqual(self.tr.selected_cds_end, None)

    def test_remove_utr(self):
        """Test for CDS stripping. We remove the UTRs and verify that start/end have moved, no UTR is present, etc.
        """

        self.tr.remove_utrs()
        self.assertEqual(self.tr.selected_cds_start, self.tr.start)
        self.assertEqual(self.tr.selected_cds_end, self.tr.end)
        self.assertEqual(self.tr.three_utr, [])
        self.assertEqual(self.tr.five_utr, [])
        self.assertEqual(self.tr.combined_cds,
                         [tuple([8571, 8666])],
                         self.tr.combined_cds)
        self.assertEqual(self.tr.combined_utr, [], self.tr.combined_utr)

    def test_negative_orf(self):
        """Test loading a negative strand ORF onto a monoexonic transcript.
        This should reverse the ORF."""

        self.orf.strand = "-"
        self.tr.strip_cds(strand_specific=False)
        self.orf.has_stop_codon = False
        self.tr.load_orfs([self.orf])
        self.assertEqual(self.tr.strand, "-")
        self.assertEqual(self.tr.selected_cds_start, 8666)
        self.assertEqual(self.tr.selected_cds_end, 8571)

    def test_introns(self):

        self.assertEqual(self.tr.introns,
                         set([
                         ]),
                         self.tr.introns
                         )
        self.assertEqual(self.tr.combined_cds_introns,
                         set([
                         ]),
                         self.tr.combined_cds_introns
                         )
        self.assertEqual(self.tr.selected_cds_introns,
                         set([
                         ]),
                         self.tr.selected_cds_introns
                         )

    def testDoubleOrf(self):

        """Test to verify the introduction of multiple ORFs."""

        self.tr.strip_cds()
        self.tr.finalized = False

        first_orf = parsers.bed12.BED12()
        first_orf.chrom = self.tr.id
        first_orf.start = 1
        first_orf.end = self.tr.cdna_length
        first_orf.name = self.tr.id
        first_orf.strand = "+"
        first_orf.score = 0
        first_orf.thick_start = 51
        first_orf.thick_end = 398
        first_orf.block_count = 1
        first_orf.blockSize = self.tr.cdna_length
        first_orf.block_sizes = [self.tr.cdna_length]
        first_orf.block_starts = [0]
        first_orf.rgb = 0
        first_orf.has_start_codon = True
        first_orf.has_stop_codon = True
        first_orf.transcriptomic = True
        self.assertFalse(first_orf.invalid)

        # This should not be incorporated
        second_orf = parsers.bed12.BED12()
        second_orf.chrom = self.tr.id
        second_orf.start = 1
        second_orf.end = self.tr.cdna_length
        second_orf.name = "second"
        second_orf.strand = "+"
        second_orf.score = 0
        second_orf.thick_start = 201
        second_orf.thick_end = 410
        second_orf.block_count = 1
        second_orf.blockSize = self.tr.cdna_length
        second_orf.block_sizes = [self.tr.cdna_length]
        second_orf.block_starts = [0]
        second_orf.rgb = 0
        second_orf.has_start_codon = True
        second_orf.has_stop_codon = True
        second_orf.transcriptomic = True
        self.assertFalse(second_orf.invalid)

        self.assertTrue(loci.Transcript.is_overlapping_cds(
            first_orf, second_orf))

        # This should be added
        third_orf = parsers.bed12.BED12()
        third_orf.chrom = self.tr.id
        third_orf.start = 1
        third_orf.end = self.tr.cdna_length
        third_orf.name = "third"
        third_orf.strand = "+"
        third_orf.score = 0
        third_orf.thick_start = 501
        third_orf.thick_end = 800
        third_orf.block_count = 1
        third_orf.blockSize = self.tr.cdna_length
        third_orf.block_sizes = [self.tr.cdna_length]
        third_orf.block_starts = [0]
        third_orf.rgb = 0
        third_orf.has_start_codon = True
        third_orf.has_stop_codon = True
        third_orf.transcriptomic = True
        self.assertFalse(third_orf.invalid)

        self.assertFalse(loci.Transcript.is_overlapping_cds(first_orf, third_orf))
        self.assertFalse(loci.Transcript.is_overlapping_cds(second_orf, third_orf))

        self.assertFalse(third_orf == second_orf)
        self.assertFalse(first_orf == second_orf)
        self.assertFalse(first_orf == third_orf)

        candidates = [first_orf, second_orf, third_orf]

        self.tr.logger = self.logger

        self.tr.load_orfs([first_orf])
        self.tr.load_orfs([second_orf])
        self.tr.load_orfs([third_orf])

        self.tr.load_orfs([first_orf, second_orf, third_orf])

        self.assertTrue(self.tr.is_complete)
        self.tr.finalize()
        self.assertEqual(self.tr.number_internal_orfs, 2,
                         (self.tr.cdna_length, self.tr.selected_start_distance_from_tss,
                          self.tr.selected_end_distance_from_tes))

        self.assertEqual(self.tr.combined_cds_length, 648)
        self.assertEqual(self.tr.selected_cds_length, 348)
        self.assertEqual(self.tr.number_internal_orfs, 2,
                         "\n".join([str(x) for x in self.tr.internal_orfs]))

        new_transcripts = sorted(self.tr.split_by_cds())

        self.assertEqual(len(new_transcripts), 2)
        self.assertEqual(new_transcripts[0].three_utr_length, 0)
        self.assertEqual(new_transcripts[1].five_utr_length, 0)

    def testDoubleOrf_negative(self):

        """Test to verify the introduction of multiple ORFs."""

        self.tr.strip_cds(strand_specific=False)
        self.tr.finalized = False

        first_orf = parsers.bed12.BED12()
        first_orf.chrom = self.tr.id
        first_orf.start = 1
        first_orf.end = self.tr.cdna_length
        first_orf.name = self.tr.id
        first_orf.strand = "-"
        first_orf.score = 0
        first_orf.thick_start = 51
        first_orf.thick_end = 398
        first_orf.block_count = 1
        first_orf.blockSize = self.tr.cdna_length
        first_orf.block_sizes = [self.tr.cdna_length]
        first_orf.block_starts = [0]
        first_orf.rgb = 0
        first_orf.has_start_codon = True
        first_orf.has_stop_codon = True
        first_orf.transcriptomic = True
        self.assertFalse(first_orf.invalid)

        # This should not be incorporated
        second_orf = parsers.bed12.BED12()
        second_orf.chrom = self.tr.id
        second_orf.start = 1
        second_orf.end = self.tr.cdna_length
        second_orf.name = "second"
        second_orf.strand = "-"
        second_orf.score = 0
        second_orf.thick_start = 201
        second_orf.thick_end = 410
        second_orf.block_count = 1
        second_orf.blockSize = self.tr.cdna_length
        second_orf.block_sizes = [self.tr.cdna_length]
        second_orf.block_starts = [0]
        second_orf.rgb = 0
        second_orf.has_start_codon = True
        second_orf.has_stop_codon = True
        second_orf.transcriptomic = True

        self.assertFalse(second_orf.invalid)

        # self.assertTrue(loci.Transcript.is_overlapping_cds(first_orf,
        #                                                           second_orf))

        # This should be added
        third_orf = parsers.bed12.BED12()
        third_orf.chrom = self.tr.id
        third_orf.start = 1
        third_orf.end = self.tr.cdna_length
        third_orf.name = "third"
        third_orf.strand = "-"
        third_orf.score = 0
        third_orf.thick_start = 501
        third_orf.thick_end = 800
        third_orf.block_count = 1
        third_orf.blockSize = self.tr.cdna_length
        third_orf.block_sizes = [self.tr.cdna_length]
        third_orf.block_starts = [0]
        third_orf.rgb = 0
        third_orf.has_start_codon = True
        third_orf.has_stop_codon = True
        third_orf.transcriptomic = True
        self.assertFalse(third_orf.invalid)

        self.assertFalse(
            loci.Transcript.is_overlapping_cds(
                first_orf, third_orf))
        self.assertFalse(
            loci.Transcript.is_overlapping_cds(
                second_orf, third_orf))

        self.assertFalse(third_orf == second_orf)
        self.assertFalse(first_orf == second_orf)
        self.assertFalse(first_orf == third_orf)

        candidates = [first_orf, second_orf, third_orf]

        # self.assertEqual(len(self.tr.find_overlapping_cds(candidates)), 2)

        self.tr.logger = self.logger

        self.tr.load_orfs(candidates)

        self.assertTrue(self.tr.is_complete)
        self.tr.finalize()
        self.assertEqual(self.tr.number_internal_orfs, 2, (
            self.tr.cdna_length,
            self.tr.selected_start_distance_from_tss,
            self.tr.selected_end_distance_from_tes))

        # self.assertEqual(self.tr.combined_cds_length, 648)
        self.assertEqual(self.tr.selected_cds_length, 348)
        self.assertEqual(self.tr.number_internal_orfs, 2, "\n".join([str(x) for x in self.tr.internal_orfs]))

        new_transcripts = sorted(self.tr.split_by_cds())

        self.assertEqual(len(new_transcripts), 2)
        self.assertEqual(new_transcripts[0].five_utr_length, 0)
        self.assertEqual(new_transcripts[1].three_utr_length, 0)

    def test_wrong_orf(self):
        # This should be added
        orf = parsers.bed12.BED12()
        orf.chrom = self.tr.id
        orf.start = 1
        orf.end = self.tr.cdna_length + 1
        orf.name = "third"
        orf.strand = "-"
        orf.score = 0
        orf.thick_start = 501
        orf.thick_end = 800
        orf.block_count = 1
        orf.blockSize = self.tr.cdna_length
        orf.block_sizes = [self.tr.cdna_length]
        orf.block_starts = [0]
        orf.rgb = 0
        orf.has_start_codon = True
        orf.has_stop_codon = True
        orf.transcriptomic = True
        self.assertFalse(orf.invalid)

        self.tr.logger = self.logger
        self.tr.strip_cds()
        self.tr.strand = "+"
        self.logger.setLevel("WARNING")
        # self.tr.load_orfs([orf])
        with self.assertLogs("null", level="DEBUG") as cm_out:
            self.tr.load_orfs([orf])

        self.assertFalse(self.tr.is_coding)


class TestWheatRNA(unittest.TestCase):

    def test_negative_orf(self):

        transcript = loci.Transcript()
        transcript.chrom = "Triticum_aestivum_CS42_TGACv1_scaffold_018953_1AS"
        transcript.strand = None
        transcript.start = 215963
        transcript.end = 217518
        transcript.id = "TGAC_Root_Cufflinks_CL_Root.3672.1"
        transcript.add_exons([(215963, 217518)])
        transcript.parent = "foo"
        transcript.score = 23
        transcript.finalize()

        bed = parsers.bed12.BED12()
        bed.chrom = "TGAC_Root_Cufflinks_CL_Root.3672.1"
        bed.transcriptomic = True
        bed.start = 1
        bed.end = 1556
        bed.name = "ID=TGAC_Root_Cufflinks_CL_Root.3672.1|m.294;TGAC_Root_Cufflinks_CL_Root.3672.1|g.294;ORF_TGAC_Root_Cufflinks_CL_Root.3672.1|g.294_TGAC_Root_Cufflinks_CL_Root.3672.1|m.294_type:5prime_partial_len:157_(-)"
        bed.strand = "-"
        bed.thick_start = 1084
        bed.thick_end = 1554
        bed.has_start_codon = False
        bed.has_stop_codon = True
        bed.header = False
        self.assertFalse(bed.invalid)

        bed2 = parsers.bed12.BED12()
        bed2.chrom = "TGAC_Root_Cufflinks_CL_Root.3672.1"
        bed2.transcriptomic = True
        bed2.start = 1
        bed2.end = 1556
        bed2.name = "ID=TGAC_Root_Cufflinks_CL_Root.3672.1|name2"
        bed2.strand = "-"
        bed2.thick_start = 884
        bed2.thick_end = 1054
        bed2.has_start_codon = False
        bed2.has_stop_codon = False
        bed2.header = False
        self.assertFalse(bed2.invalid)

        transcript.logger = create_default_logger("test_wheat_negative_orf", "DEBUG")
        transcript.configuration.pick.orf_loading.minimal_secondary_orf_length = 150
        transcript.load_orfs([bed, bed2])
        self.assertEqual(len(transcript.internal_orfs), 2)
        transcript.finalize()
        self.assertEqual(transcript.start, 215963)
        self.assertEqual(transcript.end, 217518)
        self.assertEqual(transcript.strand, "-")
        self.assertEqual((transcript.start,
                          transcript.selected_cds_start,
                          transcript.selected_cds_end,
                          transcript.end),
                         (215963,
                          217516,
                          217046,
                          217518
                          ))

    def test_orf_sorter(self):
        bed = parsers.bed12.BED12()
        bed.chrom = "TGAC_Root_Cufflinks_CL_Root.3672.1"
        bed.transcriptomic = True
        bed.start = 1
        bed.end = 1556
        bed.name = "ID=TGAC_Root_Cufflinks_CL_Root.3672.1|m.294;TGAC_Root_Cufflinks_CL_Root.3672.1|g.294;ORF_TGAC_Root_Cufflinks_CL_Root.3672.1|g.294_TGAC_Root_Cufflinks_CL_Root.3672.1|m.294_type:5prime_partial_len:157_(-)"
        bed.strand = "-"
        bed.thick_start = 1084
        bed.thick_end = 1554
        bed.has_start_codon = False
        bed.has_stop_codon = True
        self.assertFalse(bed.invalid)

        bed2 = parsers.bed12.BED12()
        bed2.chrom = "TGAC_Root_Cufflinks_CL_Root.3672.1"
        bed2.transcriptomic = True
        bed2.start = 1
        bed2.end = 1556
        bed2.name = "ID=TGAC_Root_Cufflinks_CL_Root.3672.1|2"
        bed2.strand = "-"
        bed2.thick_start = 884
        bed2.thick_end = 1054
        bed2.has_start_codon = False
        bed2.has_stop_codon = False

        after_sorting = sorted([bed, bed2],
                               reverse=True,
                               key=transcripts.transcript_methods.retrieval.orf_sorter
                               )

        self.assertEqual(after_sorting[0], bed)


class TestNegativeSplit(unittest.TestCase):

    def setUp(self):

        tr = """Triticum_aestivum_CS42_TGACv1_scaffold_018974_1AS\tCufflinks\ttranscript\t72914\t76276\t1000\t.\t.\tgene_id "CL_10DPA.2184"; transcript_id "ERP004505_10DPA_Cufflinks_CL_10DPA.2184.1"; exon_number "1"; FPKM "0.1596439944"; conf_hi "0.205257"; conf_lo "0.114031"; frac "1.000000"; cov "3.066911"; Name "ERP004505_10DPA_Cufflinks_CL_10DPA.2184.1";
Triticum_aestivum_CS42_TGACv1_scaffold_018974_1AS\tCufflinks\texon\t72914\t76276\t.\t.\t.\tgene_id "CL_10DPA.2184"; transcript_id "ERP004505_10DPA_Cufflinks_CL_10DPA.2184.1";"""
        tr_lines = [parsers.GTF.GtfLine(_) for _ in tr.split("\n")]
        self.tr = loci.Transcript(tr_lines[0])
        self.tr.add_exon(tr_lines[1])

        self.bed1 = parsers.bed12.BED12()
        self.bed1.header = False
        self.bed1.chrom = "ERP004505_10DPA_Cufflinks_CL_10DPA.2184.1"
        self.bed1.start = 1
        self.bed1.end = 3363
        self.bed1.name = "First|m.2472_type:complete_len:193_(-)"
        self.bed1.strand = "-"
        self.bed1.thick_start = 1423
        self.bed1.thick_end = 2001
        self.bed1.has_stop_codon = True
        self.bed1.has_start_codon = True
        self.bed1.transcriptomic = True
        self.assertEqual(self.bed1.cds_len, 579)
        
        self.bed2 = parsers.bed12.BED12()
        self.bed2.header = False
        self.bed2.chrom = "ERP004505_10DPA_Cufflinks_CL_10DPA.2184.1"
        self.bed2.start = 1
        self.bed2.end = 3363
        self.bed2.name = "Second|m.2472_type:complete_len:137_(-)"
        self.bed2.strand = "-"
        self.bed2.thick_start = 2481
        self.bed2.thick_end = 2891
        self.bed2.has_stop_codon = True
        self.bed2.has_start_codon = True
        self.bed2.transcriptomic = True
        self.assertEqual(self.bed2.cds_len, 411)

    def test_loading(self):
        
        self.tr.load_orfs([self.bed1, self.bed2])
        self.assertEqual(self.tr.number_internal_orfs, 2)
        self.assertEqual(self.tr.selected_cds_start,
                         74914)
        self.assertEqual(self.tr.selected_cds_end,
                         74336)
        self.assertEqual(self.tr.combined_cds_end,
                         74336)
        self.assertEqual(self.tr.combined_cds_start,
                         75804)

    def test_split(self):
        self.tr.load_orfs([self.bed1, self.bed2])
        self.tr.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"
        self.assertEqual(self.tr.number_internal_orfs, 2)
        logger = create_default_logger("splitter")
        logger.setLevel("ERROR")
        self.tr.logger = logger

        new_transcripts = [_ for _ in self.tr.split_by_cds()]

        new_transcripts = sorted(new_transcripts,
                                 key=operator.attrgetter("start", "end"))
        self.assertEqual(len(new_transcripts), 2)

        self.assertEqual(new_transcripts[0].start, 72914)
        self.assertEqual(new_transcripts[0].end, 74914)

        self.assertEqual(new_transcripts[1].end, 76276, self.tr.internal_orfs)
        self.assertEqual(new_transcripts[1].start, 75394)

        self.assertEqual(new_transcripts[0].selected_cds_start, 74914)
        self.assertEqual(new_transcripts[0].selected_cds_end, 74336)

        self.assertEqual(new_transcripts[1].selected_cds_start, 75804)
        self.assertEqual(new_transcripts[1].selected_cds_end, 75394)


class TestTripleNegative(unittest.TestCase):

    def setUp(self):

        trlines = """Triticum_aestivum_CS42_TGACv1_scaffold_019715_1AS\tCufflinks\ttranscript\t44187\t49369\t1000\t.\t.\tgene_id "CL_Spike.6733"; transcript_id "TGAC_Spike_Cufflinks_CL_Spike.6733.1"; exon_number "1"; cov "2963.838405"; frac "0.338459"; FPKM "46.8523047496"; conf_hi "48.870099"; Name "TGAC_Spike_Cufflinks_CL_Spike.6733.1"; conf_lo "44.834511";
Triticum_aestivum_CS42_TGACv1_scaffold_019715_1AS\tCufflinks\texon\t44187\t49369\t.\t.\t.\tgene_id "CL_Spike.6733"; transcript_id "TGAC_Spike_Cufflinks_CL_Spike.6733.1";"""
        trlines = [parsers.GTF.GtfLine(_) for _ in trlines.split("\n")]

        self.tr = loci.Transcript(trlines[0])
        self.tr.add_exon(trlines[1])
        self.tr.finalize()

        self.bed1 = parsers.bed12.BED12(transcriptomic=True)
        self.bed1.header = False
        self.bed1.chrom = self.tr.id
        self.bed1.start = 1
        self.bed1.end = 5183
        self.bed1.strand = "-"
        self.bed1.name = "g.132606_TGAC_Spike_Cufflinks_CL_Spike.6733.1|m.132606_type:3prime_partial_len:144_(-)"
        self.bed1.thick_start = 2
        self.bed1.thick_end = 430
        self.bed1.score = 0
        # self.bed1.has_start_codon = True
        # self.bed1.has_stop_codon = False
        self.assertEqual(self.bed1.cds_len, 429)
        
        self.bed2 = parsers.bed12.BED12(transcriptomic=True)
        self.bed2.header = False
        self.bed2.chrom = self.tr.id
        self.bed2.start = 1
        self.bed2.end = 5183
        self.bed2.strand = "-"
        self.bed2.name = "g.132602_TGAC_Spike_Cufflinks_CL_Spike.6733.1|m.132602_type:complete_len:387_(-)"
        self.bed2.thick_start = 936
        self.bed2.thick_end = 2096
        self.bed2.score = 0
        # self.bed2.has_stop_codon = True
        # self.bed2.has_start_codon = True
        self.assertEqual(self.bed2.cds_len, 1161)
        
        self.bed3 = parsers.bed12.BED12(transcriptomic=True)
        self.bed3.header = False
        self.bed3.chrom = self.tr.id
        self.bed3.start = 1
        self.bed3.end = 5183
        self.bed3.strand = "-"
        self.bed3.name = "g.132601_TGAC_Spike_Cufflinks_CL_Spike.6733.1|m.132601_type:5prime_partial_len:542_(-)"
        # self.bed3.has_start_codon = False
        # self.bed3.has_stop_codon = True
        self.bed3.thick_start = 3558
        self.bed3.thick_end = 5183
        self.bed3.score = 0
        self.assertEqual(self.bed3.cds_len, 1626)

    def test_split(self):

        self.tr.load_orfs([self.bed1, self.bed2, self.bed3])
        self.tr.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        new = sorted([_ for _ in self.tr.split_by_cds()], key=operator.attrgetter("start", "end"))
        self.assertEqual(len(new), 3)
        self.assertEqual(new[0].start, 44187)
        self.assertEqual(new[0].end, 44616)
        self.assertEqual(new[0].selected_cds_end, 44188)

        self.assertEqual(new[1].start, new[1].selected_cds_end)
        self.assertEqual(new[1].start, 45122)
        self.assertEqual(new[1].end, new[1].selected_cds_start)
        self.assertEqual(new[1].end, 46282)

        self.assertEqual(new[2].start, 47744)
        self.assertEqual(new[2].end, self.tr.end)
        self.assertEqual(new[2].end, new[2].selected_cds_start)


class TestNegativeMonoexonicLoad(unittest.TestCase):

    def setUp(self):

        trlines="""Chr3\tCufflinks\ttranscript\t2949168\t2952410\t1000\t-\t.\tgene_id "cufflinks_star_at.10687"; transcript_id "cufflinks_cufflinks_star_at.10687.1"; exon_number "1"; conf_lo "0.169545"; FPKM "0.1111324743"; frac "0.781250"; conf_hi "0.308263"; cov "3.129756";
Chr3\tCufflinks\texon\t2949168\t2952410\t.\t-\t.\tgene_id "cufflinks_star_at.10687"; transcript_id "cufflinks_cufflinks_star_at.10687.1";"""

        trlines = [parsers.GTF.GtfLine(_) for _ in trlines.split("\n")]

        self.tr = loci.Transcript(trlines[0])
        self.tr.add_exon(trlines[1])
        self.tr.finalize()

        self.bed1 = parsers.bed12.BED12()
        self.bed1.header = False
        self.bed1.chrom = self.tr.id
        self.bed1.start = 1
        self.bed1.end = 3243
        self.bed1.transcriptomic = True
        self.bed1.thick_start = 203
        self.bed1.thick_end = 2206
        self.bed1.name = "g.135538_cufflinks_cufflinks_star_at.10687.1|m.135538_type:complete_len:668_(+)"
        self.bed1.has_start_codon = True
        self.bed1.has_stop_codon = True
        self.bed1.strand = "+"
        self.assertEqual(self.bed1.cds_len, 2004)
        self.bed1.score = 0
        self.assertFalse(self.bed1.invalid)

        self.bed2 = parsers.bed12.BED12()
        self.bed2.header = False
        self.bed2.chrom = self.tr.id
        self.bed2.start = 1
        self.bed2.end = 3243
        self.bed2.transcriptomic = True
        self.bed2.thick_start = 2543
        self.bed2.thick_end = 3241
        self.bed2.strand = "+"
        self.bed2.name = "g.135539_cufflinks_cufflinks_star_at.10687.1|m.135539_type:3prime_partial_len:234_(+)"
        self.bed2.has_start_codon = True
        self.bed2.has_stop_codon = False
        self.assertEqual(self.bed2.cds_len, 699)
        self.bed2.score = 0
        self.assertFalse(self.bed2.invalid)

    def test_load(self):

        self.tr.load_orfs([self.bed1, self.bed2])
        self.assertEqual(self.tr.strand, "-")

        self.assertEqual(self.tr.number_internal_orfs, 2)
        self.assertEqual(self.tr.combined_cds_end, 2949170)
        self.assertEqual(self.tr.combined_cds_start, 2952208)
        self.assertEqual(self.tr.selected_cds_start, 2952208)
        self.assertEqual(self.tr.selected_cds_end, 2950205)

    # @unittest.skip
    def test_split(self):
# Chr3	Cufflinks	mRNA	2949168	2952410	1000	-	.	ID=cufflinks_cufflinks_star_at.10687.1.orf1;
# Chr3	Cufflinks	three_prime_UTR	2949168	2950204	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf1
# Chr3	Cufflinks	exon	2949168	2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf1
# Chr3	Cufflinks	CDS	2,950,205	2,952,208	.	-	0	Parent=cufflinks_cufflinks_star_at.10687.1.orf1
# Chr3	Cufflinks	five_prime_UTR	2952209	2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf1
#
# Chr3	Cufflinks	mRNA	2949168	2952410	1000	-	.	ID=cufflinks_cufflinks_star_at.10687.1.orf2;
# Chr3	Cufflinks	three_prime_UTR	2949168	2949169	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf2
# Chr3	Cufflinks	exon	2949168	2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf2
# Chr3	Cufflinks	CDS	2,949,170	2,949,868	.	-	0	Parent=cufflinks_cufflinks_star_at.10687.1.orf2
# Chr3	Cufflinks	five_prime_UTR	2949869	2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.orf2
# =============>
# Chr3	Cufflinks	mRNA	2950205	2952410	1000	-	.	ID=cufflinks_cufflinks_star_at.10687.1.split1;
# Chr3	Cufflinks	exon	2950205 2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.split1
# Chr3	Cufflinks	CDS	2950205	2952208	.	-	0	Parent=cufflinks_cufflinks_star_at.10687.1.split1
# Chr3	Cufflinks	five_prime_UTR	2952209	2952410	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.split1
#
# Chr3	Cufflinks	mRNA	2949168	2949868	1000	-	.	ID=cufflinks_cufflinks_star_at.10687.1.split2;
# Chr3	Cufflinks	three_prime_UTR	2949168	2949169	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.split2
# Chr3	Cufflinks	exon	2949168	2949868	.	-	.	Parent=cufflinks_cufflinks_star_at.10687.1.split2
# Chr3	Cufflinks	CDS	2949170	2949868	.	-	0	Parent=cufflinks_cufflinks_star_at.10687.1.split2

        self.tr.load_orfs([self.bed1, self.bed2])
        self.assertEqual(self.tr.strand, "-")

        self.assertEqual(self.tr.number_internal_orfs, 2)
        self.assertEqual(self.tr.combined_cds_end, 2949170)
        self.assertEqual(self.tr.combined_cds_start, 2952208)
        self.assertEqual(self.tr.selected_cds_start, 2952208)
        self.assertEqual(self.tr.selected_cds_end, 2950205)

        logger = create_default_logger("splitter")
        logger.setLevel("DEBUG")
        self.tr.logger = logger
        self.tr.configuration.pick.chimera_split.blast_params.leniency = "LENIENT"

        new_transcripts = sorted([_ for _ in self.tr.split_by_cds()])

        self.assertEqual(new_transcripts[0].start, self.tr.start)
        self.assertEqual(new_transcripts[0].end, 2949868, "\n\n".join([str(_) for _ in new_transcripts]))


if __name__ == '__main__':
    unittest.main()



"""
This package contains all the modules necessary for BLAST serialisation and analysis.
"""

from ...exceptions import InvalidHit
from .btop_parser import parse_btop
from .query import Query
from .target import Target
from .hsp import Hsp, prepare_hsp
from .hit import Hit, prepare_hit
from .blast_serialiser import BlastSerializer


__author__ = 'Luca Venturini'



#!/usr/bin/env python3

import os
import argparse


class Stats:

    file = ""
    seqs = 0
    mapped = 0
    mapped_n_paired = 0
    unmapped = 0
    properly_paired = 0
    paired = 0
    duplicated = 0
    MQ0 = 0

    def __init__(self):
        self.data = []

    def __str__(self):

        return "\t".join([str(_) for _ in [self.file,
                                           self.seqs,
                                           self.mapped,
                                           self.mapped_n_paired,
                                           self.unmapped,
                                           self.properly_paired,
                                           self.paired,
                                           self.duplicated,
                                           self.MQ0
                                           ]])

    @staticmethod
    def header():
        return "\t".join(["File",
                          "sequences",
                          "reads_mapped",
                          "reads_mapped_and_paired",
                          "reads_unmapped",
                          "reads_properly_paired",
                          "reads_paired",
                          "reads_duplicated",
                          "reads_MQ0"])


def main():
    parser = argparse.ArgumentParser("Script to collect info from multiple samtools stats files")
    parser.add_argument("input", nargs='+', help="The list of samtools stats file to process")
    args = parser.parse_args()

    print(Stats.header())

    for stats_file in args.input:
        with open(stats_file) as f:
            s = Stats()
            s.file = os.path.split(stats_file)[1]
            for line in f:
                if line.startswith("SN\t"):
                    parts = line.split("\t")
                    if parts[1] == "sequences:":
                        s.seqs = int(parts[2])
                    elif parts[1] == "reads mapped:":
                        s.mapped = int(parts[2])
                    elif parts[1] == "reads mapped and paired:":
                        s.mapped_n_paired = int(parts[2])
                    elif parts[1] == "reads unmapped:":
                        s.unmapped = int(parts[2])
                    elif parts[1] == "reads properly paired:":
                        s.properly_paired = int(parts[2])
                    elif parts[1] == "reads paired:":
                        s.paired = int(parts[2])
                    elif parts[1] == "reads duplicated:":
                        s.duplicated = int(parts[2])
                    elif parts[1] == "reads MQ0:":
                        s.MQ0 = int(parts[2])
            print(s)


main()



class Namespace:

    __name__ = "Namespace"

    def __init__(self, default=0, **kwargs):
        self.__default = default
        self.__values = dict()
        self.update(kwargs)

    @property
    def default(self):
        return self.__default

    def __getitem__(self, item):
        self.__dict__.setdefault(item, self.default)
        self.__values[item] = self.__dict__[item]
        return self.__values[item]

    def __setitem__(self, key, value):
        self.__dict__[key] =  value
        self.__values[key] = value

    def __getstate__(self):

        default = self.default
        del self.__default
        state = dict()
        state.update(self.__dict__)
        state["default"] = default
        self.__default = default
        return state

    def __setstate__(self, state):
        self.__default = state["default"]
        del state["default"]
        self.__dict__.update(state)

    def __getattr__(self, item):
        self.__dict__.setdefault(item, self.default)
        self.__values[item] = self.__dict__[item]
        return self.__values[item]

    # DO NOT IMPLEMENT THIS __setattr__ METHOD, IT LEADS TO ERRORS IN MULTIPROCESSING (SPAWN)
    # def __setattr__(self, key, value)

    def update(self, dictionary):
        self.__dict__.update(dictionary)
        self.__values.update(dictionary)

    def get(self, item):
        return self.__getitem__(item)

    def __iter__(self):
        return iter(_ for _ in self.__values)

    def __deepcopy__(self, memodict={}):
        new = Namespace(default=self.default)
        new.update(self.__values)
        return new

    def keys(self):
        return self.__values.keys()

    def items(self):
        return self.__values.items()




import copy
import dataclasses
from dataclasses import field
import random
from marshmallow import validate, ValidationError
from marshmallow_dataclass import dataclass, Optional
from .picking_config import PickConfiguration
from .prepare_config import PrepareConfiguration
from .serialise_config import SerialiseConfiguration
from ..utilities.dbutils import DBConfiguration
from ..utilities.log_utils import LoggingConfiguration, create_null_logger, create_default_logger
from Mikado._transcripts.scoring_configuration import ScoringFile
import os
import yaml
import toml
import json
from ..exceptions import InvalidConfiguration
from pkg_resources import resource_filename
try:
    from yaml import CSafeLoader as yLoader
except ImportError:
    from yaml import SafeLoader as yLoader


@dataclass
class ReferenceConfiguration:
    genome: str = field(default="")
    genome_fai: str = field(default="")
    transcriptome: str = field(default="")


@dataclass
class MikadoConfiguration:
    """
    Configuration properties for Mikado.
    """

    scoring: Optional[ScoringFile] = field(default=None)
    threads: int = field(default=1, metadata={
        "metadata": {"description": "Threads to be used per process"},
        "validate": validate.Range(min=1),
        "required": True
    })
    seed: int = field(default=0, metadata={
        "metadata": {"description": "Random number generator seed, to ensure reproducibility across runs. Set to None"
                     "('null' in YAML/JSON/TOML files) to let Mikado select a random seed every time."},
        "validate": validate.Range(min=0, max=2 ** 32 - 1),
        "allow_none": True, "required": True
    })
    multiprocessing_method: Optional[str] = field(default="spawn", metadata={
        "metadata": {"description": "Which method (fork, spawn, forkserver) Mikado should use for multiprocessing"},
        "validate": validate.OneOf(["spawn", "fork", "fork-server"])
    })
    log_settings: LoggingConfiguration = field(default_factory=LoggingConfiguration, metadata={
                "metadata": {"description": "Settings related to the verbosity of logs"}
    })
    db_settings: Optional[DBConfiguration] = field(default_factory=DBConfiguration, metadata={
                "metadata": {"description": "Settings related to DB connection"}
    })
    serialise: SerialiseConfiguration = field(default_factory=SerialiseConfiguration, metadata={
                "metadata": {"description": "Settings related to data serialisation"}
    })
    prepare: PrepareConfiguration = field(default_factory=PrepareConfiguration, metadata={
                "metadata": {"description": "Settings related to the input data preparation"},
    })
    pick: PickConfiguration = field(default_factory=PickConfiguration, metadata={
                "metadata": {"description": "Settings related to the Mikado pick stage"},
    })
    reference: Optional[ReferenceConfiguration] = field(default_factory=ReferenceConfiguration, metadata={
            })

    _loaded_scoring = None
    filename: Optional[str] = field(default=None)

    def __post_init__(self):
        self.check()

    def copy(self):
        return copy.copy(self)

    def check(self, logger=create_null_logger()):
        if self.seed is None:
            self.seed = random.randint(0, 2 ** 32 - 1)
            logger.info(f"Random seed: {self.seed}")
        if self.scoring is None or not hasattr(self.scoring.requirements, "parameters"):
            self.load_scoring(logger=logger)
        self.scoring.check(minimal_orf_length=self.pick.orf_loading.minimal_orf_length)
        errors = self.Schema().validate(dataclasses.asdict(self))
        if len(errors) > 0:
            exc = InvalidConfiguration(f"The configuration is invalid, please double check. Errors:\n{errors}")
            logger.critical(exc)
            raise exc

    def load_scoring(self, logger=None):
        """
        The purpose of this section is the following:
        - check that the scoring file exists somewhere different from the system folder. If it does, check whether it is
          a valid file.
        - If the local file is broken / wrong / inexistent, use the system one.
        - load the scoring file into the configuration object.

        :param configuration: configuration object to check
        :param logger: logger to use
        """

        if logger is None:
            logger = create_default_logger("check_scoring", level="WARNING")
        if self.pick.scoring_file is None:
            if self._loaded_scoring != self.pick.scoring_file:            
                logger.warning(f"Resetting the scoring to its previous value ({self._loaded_scoring})")
                self.pick.scoring_file = self._loaded_scoring = os.path.abspath(self._loaded_scoring)
        elif self._loaded_scoring != self.pick.scoring_file:
            logger.debug("Overwriting the scoring self using '%s' as scoring file", self.pick.scoring_file)
            self.scoring_file = None
        else:
            logger.debug("Restarting")

        options = [os.path.abspath(self.pick.scoring_file),
                   os.path.abspath(os.path.join(os.path.dirname(self.filename or ""),
                                                self.pick.scoring_file)),
                   os.path.abspath(os.path.join(resource_filename("Mikado.configuration", "scoring_files"),
                                                self.pick.scoring_file))]

        if self.filename is not None:
            options.append(os.path.join(os.path.dirname(os.path.abspath(self.filename)),
                                        os.path.basename(self.pick.scoring_file)))
            if not os.path.isabs(self.pick.scoring_file):
                options.append(os.path.join(os.path.dirname(os.path.abspath(self.filename)),
                                            self.pick.scoring_file))

        found = False

        for option in options:
            if not os.path.exists(option):
                continue
            if option.endswith(("yaml", "json", "toml")):
                with open(option) as scoring_file:
                    if option.endswith("yaml"):
                        scoring = yaml.load(scoring_file, Loader=yLoader)
                    elif option.endswith("toml"):
                        scoring = toml.load(scoring_file)
                    else:
                        scoring = json.loads(scoring_file.read())
                    if not isinstance(scoring, dict):
                        continue
                    try:
                        checked = ScoringFile.Schema().load(scoring)
                        checked.check(minimal_orf_length=self.pick.orf_loading.minimal_orf_length)
                        self._loaded_scoring = self.pick.scoring_file
                    except (InvalidConfiguration, ValidationError) as exc:
                        msg = f"The configuration file {option} is invalid:\n"
                        msg += str(exc)
                        logger.critical(msg)
                        raise InvalidConfiguration(msg)
                    self.scoring = checked
                    found = True
                    self.pick.scoring_file = option
            if found is True:
                logger.info("Found the correct option: %s", option)
                self.pick.scoring_file = os.path.abspath(option)
                self._loaded_scoring = os.path.abspath(option)
                break
        if not found:
            raise InvalidConfiguration("No scoring configuration file found. Options: {}".format(",".join(options)))
        return



#!/usr/bin/env python3

"""Stub of pre-configurer for Mikado"""
import yaml
import os
import re
from pkg_resources import resource_filename, resource_stream
import glob
import argparse
import sys
import dataclasses

from ._utils import _set_pick_mode
from ..configuration import DaijinConfiguration, MikadoConfiguration
from ..exceptions import InvalidConfiguration
from ..utilities import comma_split, percentage, merge_dictionaries
from ..utilities.namespace import Namespace
from ..configuration.configurator import load_and_validate_config
from ..configuration import print_config
import tempfile
from ..utilities.log_utils import create_null_logger, create_default_logger
try:
    from yaml import CSafeLoader as yLoader
except ImportError:
    from yaml import SafeLoader as yLoader
from .prepare import parse_prepare_options

__author__ = 'Luca Venturini'


def __add_daijin_specs(args, config):
    from ..configuration.daijin_configurator import create_cluster_config, create_daijin_config
    namespace = Namespace(default=False)
    namespace.r1 = []
    namespace.r2 = []
    namespace.samples = []
    namespace.strandedness = []
    namespace.asm_methods = []
    namespace.long_aln_methods = []
    namespace.aligners = []
    if args.mode is not None:
        namespace.modes = args.mode[:]
    else:
        namespace.modes = ["stringent"]
    if args.blast_targets:
        namespace.prot_db = args.blast_targets[:]
    else:
        namespace.prot_db = []
    namespace.cluster_config = None
    namespace.scheduler = ""
    namespace.flank = None
    namespace.intron_range = None
    namespace.genome = args.reference
    namespace.transcriptome = ""
    namespace.name = "Daijin"
    namespace.out_dir = args.out_dir if args.out_dir else "Daijin"
    namespace.threads = args.threads
    namespace.scoring = args.scoring
    namespace.new_scoring = getattr(args, "new_scoring", None)
    namespace.full = args.full
    namespace.no_files = args.no_files
    config = create_daijin_config(namespace, config, level="ERROR", piped=True)
    config.blastx.chunks = args.blast_chunks
    config.mikado.use_diamond = (not args.use_blast)
    config.mikado.use_prodigal = (not args.use_transdecoder)
    config.scheduler = args.scheduler
    create_cluster_config(config, args, create_null_logger())
    return config


def create_config(args):
    """
    Utility to create a default configuration file.
    :param args:
    :return:
    """

    if isinstance(args.seed, bool) or args.seed is None:
        raise OSError("Invalid seed: {}".format(args.seed))
    elif not (isinstance(args.seed, int) and 0 <= args.seed <= 2 ** 32 - 1):
        raise OSError("Invalid seed: {}".format(args.seed))

    args.daijin = args.daijin or [isinstance(args.mode, list) and len(args.mode) > 1]
    config = DaijinConfiguration() if args.daijin is True else MikadoConfiguration()

    if args.external is not None:
        other = dataclasses.asdict(load_and_validate_config(args.external, external=True))
        config = dataclasses.asdict(config)
        config = merge_dictionaries(config, other)
        config = load_and_validate_config(config)

    if args.daijin is True:
        config = __add_daijin_specs(args, config)

    config.pick.files.subloci_out = args.subloci_out if args.subloci_out else ""
    config.pick.files.monoloci_out = args.monoloci_out if args.monoloci_out else ""

    if isinstance(args.gff, str):
        args.gff = args.gff.split(",")
    elif not args.gff:
        args.gff = []
    config = parse_prepare_options(args, config)

    if args.random_seed is True:
        config.seed = None
    else:
        config.seed = args.seed

    config.serialise.files.junctions = args.junctions if args.junctions is not None else \
        config.serialise.files.junctions

    config.serialise.files.blast_targets = args.blast_targets if args.blast_targets is not None else \
        config.serialise.files.blast_targets

    if args.only_reference_update is True or args.reference_update is True:
        if len(config.prepare.files.reference) == 0:
            logger = create_default_logger("configure")
            logger.error(
                "No reference dataset provided! Please correct the issue or remove the \"--only-reference-update\" \
switch.")
            sys.exit(1)
        else:
            config.pick.run_options.only_reference_update = True

    config.pick.run_options.check_references = True if args.check_references is True else \
        config.pick.run_options.check_references

    config.pick.output_format.report_all_orfs = True if args.report_all_orfs is True else \
        config.pick.output_format.report_all_orfs

    config.pick.output_format.report_all_external_metrics = True if args.report_all_external_metrics else \
        config.pick.output_format.report_all_external_metrics

    if args.scoring is not None:
        if args.copy_scoring is not False:
            with open(args.copy_scoring, "wt") as out:
                with resource_stream("Mikado", os.path.join("configuration",
                                                            "scoring_files",
                                                            args.scoring)) as original:
                    for line in original:
                        print(line.decode().rstrip(), file=out)
            args.scoring = args.copy_scoring

        config.pick.scoring_file = args.scoring

    config.pick.clustering.cds_only = True if args.cds_only is True else config.pick.clustering.cds_only
    config.pick.alternative_splicing.cds_only = True if args.as_cds_only else config.pick.alternative_splicing.cds_only

    if args.daijin is False and args.mode is not None and len(args.mode) == 1:
        mode = args.mode.pop()
        config = _set_pick_mode(config, mode)

    if args.skip_split:
        if not all(_ in config.prepare.files.labels for _ in args.skip_split):
            raise InvalidConfiguration("Some of the labels to skip for splitting are invalid: {}".format(
                [_ for _ in args.skip_split if _ not in config.prepare.files.labels]
            ))
        config.pick.chimera_split.skip = list(set(config.pick.chimera_split.skip.extend(args.skip_split)))

    config.pick.alternative_splicing.pad = args.pad if args.pad is not None else config.pick.alternative_splicing.pad
    config.pick.clustering.min_cds_overlap = args.min_clustering_cds_overlap if args.min_clustering_cds_overlap is not \
        None else config.pick.clustering.min_cds_overlap

    if args.min_clustering_cdna_overlap is not None:
        config.pick.clustering.min_cdna_overlap = args.min_clustering_cdna_overlap
        if args.min_clustering_cds_overlap is None:
            config.pick.clustering.min_cds_overlap = args.min_clustering_cdna_overlap

    config.pick.run_options.intron_range = sorted(args.intron_range) if args.intron_range is not None else \
        config.pick.run_options.intron_range

    config.serialise.codon_table = str(args.codon_table) if args.codon_table not in (None, False, True) else \
        config.serialise.codon_table

    config.pick.alternative_splicing.keep_cds_disrupted_by_ri = True if args.keep_disrupted_cds is True else \
        config.pick.alternative_splicing.keep_cds_disrupted_by_ri

    config.pick.alternative_splicing.keep_retained_introns = False if args.exclude_retained_introns is True else \
        config.pick.alternative_splicing.keep_retained_introns

    if args.out_dir:
        config.prepare.files.output_dir = args.out_dir
        config.serialise.files.output_dir = args.out_dir
        config.pick.files.output_dir = args.out_dir

    config.check()

    # Check that the configuration file is correct
    with tempfile.NamedTemporaryFile("wt", suffix=".json", delete=True) as tempcheck:
        print_config(config, tempcheck, full=args.full, output_format="json")
        tempcheck.flush()
        try:
            load_and_validate_config(tempcheck.name)
        except InvalidConfiguration as exc:
            raise InvalidConfiguration("Created an invalid configuration file! Error:\n{}".format(exc))

    # Print out the final configuration file
    if args.json is True or args.out.name.endswith("json"):
        format_name = "json"
    elif args.yaml is True or args.out.name.endswith("yaml"):
        format_name = "yaml"
    else:
        format_name = "toml"

    print_config(config, args.out, output_format=format_name, no_files=args.no_files, full=args.full)


def configure_parser():
    """
    Parser for the configuration utility.
    :return: the parser.
    :rtype: argparse.ArgumentParser
    """

    scoring_folder = resource_filename("Mikado", os.path.join("configuration", "scoring_files"))
    trailing = re.compile(r"^{}".format(os.path.sep))
    fold_path = re.compile(scoring_folder)

    scoring_files = [trailing.sub(r"", re.sub(fold_path, r"", fname))
                     for fname in glob.iglob(os.path.join(scoring_folder, "**", "*yaml"), recursive=True)]

    parser = argparse.ArgumentParser(description="Configuration utility for Mikado")
    parser.add_argument("--full", action="store_true", default=False)
    seed_group = parser.add_mutually_exclusive_group()
    seed_group.add_argument("--seed", type=int, default=0, help="Random seed number. Default: 0.")
    seed_group.add_argument("--random-seed", action="store_true", default=False,
                            help="Generate a new random seed number (instead of the default of 0)")
    preparer = parser.add_argument_group("Options related to the prepare stage.")
    preparer.add_argument("--minimum-cdna-length", default=None, type=int, dest="minimum_cdna_length",
                          help="Minimum cDNA length for transcripts.")
    preparer.add_argument("--max-intron-length", default=None, type=int, dest="max_intron_length",
                          help="Maximum intron length for transcripts.")
    preparer.add_argument("--strip-faulty-cds", default=None, action="store_true",
                          help="Flag. If set, transcripts with an incorrect CDS will be retained but \
with their CDS stripped. Default behaviour: the whole transcript will be considered invalid and discarded.")
    scoring = parser.add_argument_group("Options related to the scoring system")
    scoring.add_argument("--scoring", type=str, default=None,
                         help="Scoring file to use. Mikado provides the following:\n{}".format(
                             ",\n".join(scoring_files)))
    scoring.add_argument("--copy-scoring", default=False,
                         type=str, dest="copy_scoring",
                         help="File into which to copy the selected scoring file, for modification.")
    picking = parser.add_argument_group("Options related to the picking")
    picking.add_argument("-i", "--intron-range",
                         dest="intron_range", type=int, nargs=2,
                         default=None,
                         help="""Range into which intron lengths should fall, as a couple of integers.
                             Transcripts with intron lengths outside of this range will be penalised.
                             Default: (60, 900)""")
    picking.add_argument("--subloci-out", default="", dest="subloci_out",
                         help="Name of the optional subloci output. By default, this will not be produced.")
    picking.add_argument("--monoloci-out", default="", dest="monoloci_out",
                         help="Name of the optional monoloci output. By default, this will not be produced.")
    picking.add_argument("--no-pad", dest="pad", default=None,
                         action="store_false", help="Disable transcript padding. On by default.")
    picking.add_argument("--reference-update", dest="reference_update", default=None,
                         action="store_true",
                         help="""Flag. If switched on, Mikado will prioritise transcripts marked as reference and will \
consider any other transcipt within loci only in reference to these reference transcripts. Novel loci will still be reported.""")
    picking.add_argument("--only-reference-update", dest="only_reference_update", default=None,
                         action="store_true",
                         help="""Flag. If switched on, Mikado will only keep loci where at least one of the transcripts \
    is marked as "reference". CAUTION: if no transcript has been marked as reference, \
    the output will be completely empty!""")
    picking.add_argument("-eri", "--exclude-retained-introns", default=None, action="store_true",
                         help="""Exclude all retained intron alternative splicing events from the final output. \
Default: False. Retained intron events that do not dirsupt the CDS are kept by Mikado in the final output.""")
    picking.add_argument("-kdc", "--keep-disrupted-cds", default=None, action="store_true",
                         help="""Keep in the final output transcripts whose CDS is most probably disrupted by a \
retained intron event. Default: False. Mikado will try to detect these instances and exclude them from the \
final output.""")
    picking.add_argument("--check-references", dest="check_references", default=None,
                         action="store_true",
                         help="""Flag. If switched on, Mikado will also check reference models against the general
    transcript requirements, and will also consider them as potential fragments. This is useful in the context of e.g.
    updating an *ab-initio* results with data from RNASeq, protein alignments, etc. 
    """)
    picking.add_argument("-mco", "--min-clustering-cdna-overlap", default=None, type=percentage,
                         help="Minimum cDNA overlap between two transcripts for them to be considered part of the same \
locus during the late picking stages. \
NOTE: if --min-cds-overlap is not specified, it will be set to this value! \
Default: 20%%.")
    picking.add_argument("-mcso", "--min-clustering-cds-overlap", default=None, type=percentage,
                         help="Minimum CDS overlap between two transcripts for them to be considered part of the same \
locus during the late picking stages. \
NOTE: if not specified, and --min-cdna-overlap is specified on the command line, min-cds-overlap will be set to this value! \
Default: 20%%.")
    picking.add_argument("--report-all-orfs", default=False, action="store_true",
                         help="Boolean switch. If set to true, all ORFs will be reported, not just the primary.")
    picking.add_argument("--report-all-external-metrics", default=None,
                         action="store_true",
                         help="Boolean switch. If activated, Mikado will report all available external metrics, not just \
those requested for in the scoring configuration. This might affect speed in Minos analyses.")
    picking.add_argument("--cds-only", dest="cds_only",
                         default=None, action="store_true",
                         help=""""Flag. If set, Mikado will only look for overlap in the coding features \
    when clustering transcripts (unless one transcript is non-coding, in which case  the whole transcript will \
    be considered). Please note that Mikado will only consider the **best** ORF for this. \
    Default: False, Mikado will consider transcripts in their entirety.""")
    picking.add_argument("--as-cds-only", dest="as_cds_only", default=None, action="store_true",
                         help="""Flag. If set, Mikado will only consider the CDS to determine whether a transcript
                            is a valid alternative splicing event in a locus.""")
    parser.add_argument("--strand-specific", default=False,
                        action="store_true",
                        help="""Boolean flag indicating whether all the assemblies are strand-specific.""")
    files = parser.add_mutually_exclusive_group()
    files.add_argument("--no-files", dest="no_files",
                       help="""Remove all files-specific options from the printed configuration file.
                       Invoking the "--gff" option will disable this flag.""",
                       default=False, action="store_true")
    files.add_argument("--gff", help="Input GFF/GTF file(s), separated by comma", type=str,
                       default="")
    files.add_argument("--list", type=argparse.FileType("r"),
                        help="""Tab-delimited file containing rows with the following format:
    <file>  <label> <strandedness(def. False)> <score(optional, def. 0)> <is_reference(optional, def. False)> <exclude_redundant(optional, def. True)> <strip_cds(optional, def. False)> <skip_split(optional, def. False)>
    "strandedness", "is_reference", "exclude_redundant", "strip_cds" and "skip_split" must be boolean values (True, False)
    "score" must be a valid floating number.
    """)
    parser.add_argument("--reference", "--genome", help="Fasta genomic reference.", default=None, dest="reference")
    serialisers = parser.add_argument_group(
        "Options related to the serialisation step")
    serialisers.add_argument("--junctions", type=comma_split, default=[])
    serialisers.add_argument("-bt", "--blast_targets", type=comma_split, default=None)
    parser.add_argument("--strand-specific-assemblies", type=str, default="",
                        dest="strand_specific_assemblies",
                        help="""List of strand-specific assemblies among the inputs.""")
    parser.add_argument("--labels", type=str, default="",
                        help="""Labels to attach to the IDs of the transcripts of the input files,
        separated by comma.""")
    parser.add_argument("--codon-table", dest="codon_table", default=None,
                        help="""Codon table to use. Default: 0 (ie Standard, NCBI #1, but only ATG is considered \
    a valid start codon.""")
    parser.add_argument("--external", help="""External configuration file to overwrite/add values from.
    Parameters specified on the command line will take precedence over those present in the configuration file.""")
    daijin = parser.add_argument_group("Options related to configuring a Daijin run.")
    daijin.add_argument("--daijin", action="store_true", default=False,
                        help="Flag. If set, the configuration file will be also valid for Daijin.")
    daijin.add_argument("-bc", "--blast-chunks", type=int, default=10, dest="blast_chunks",
                        help="Number of parallel DIAMOND/BLAST jobs to run. Default: %(default)s.")
    daijin.add_argument("--use-blast", action="store_true", default=False, dest="use_blast",
                        help="Flag. If switched on, Mikado will use BLAST instead of DIAMOND.")
    daijin.add_argument("--use-transdecoder", action="store_true", default=False, dest="use_transdecoder",
                        help="Flag. If switched on, Mikado will use TransDecoder instead of Prodigal.")
    daijin.add_argument("--mode", default=["stringent"], nargs="+",
                        choices=["nosplit", "stringent", "lenient", "permissive", "split"],
                        help="""Mode(s) in which Mikado will treat transcripts with multiple ORFs.
- nosplit: keep the transcripts whole.
- stringent: split multi-orf transcripts if two consecutive ORFs have both BLAST hits
             and none of those hits is against the same target.
- lenient: split multi-orf transcripts as in stringent, and additionally, also when
           either of the ORFs lacks a BLAST hit (but not both).
- permissive: like lenient, but also split when both ORFs lack BLAST hits
- split: split multi-orf transcripts regardless of what BLAST data is available.
If multiple modes are specified, Mikado will create a Daijin-compatible configuration file.""")
    daijin.add_argument("--scheduler", default="", choices=["local", "SLURM", "LSF", "PBS"],
                        help="Scheduler to use. Default: None - ie, either execute everything on the local machine or use DRMAA to submit and control jobs (recommended).")
    daijin.add_argument("--exe", default="daijin_exe.yaml",
                         help="Configuration file for the executables.")
    daijin.add_argument("-c", "--cluster_config",
                         type=str, default=None,
                         help="Cluster configuration file to write to.")
    parser.add_argument("-t", "--threads", default=1, type=int)
    parser.add_argument("--skip-split", dest="skip_split", default=[], nargs="+",
                        help="List of labels for which splitting will be disabled (eg long reads such as PacBio)")
    output_format = parser.add_mutually_exclusive_group()
    output_format.add_argument("-j", "--json", action="store_true", default=False,
                               help="Output will be in JSON (default: inferred by filename, with TOML as fallback).")
    output_format.add_argument("-y", "--yaml", action="store_true", default=False,
                               help="Output will be in YAML (default: inferred by filename, with TOML as fallback).")
    output_format.add_argument("--toml", action="store_true", default=False,
                               help="Output will be in TOML (default: inferred by filename, with TOML as fallback).")
    parser.add_argument("-od", "--out-dir", dest="out_dir", default=None,
                        help="Destination directory for the output.")
    parser.add_argument("out", nargs='?', default=sys.stdout, type=argparse.FileType('w'))
    parser.set_defaults(func=create_config)
    return parser



#!/usr/bin/env python3
# coding: utf-8

"""
Superlocus module. The class here defined is the uppermost container for transcripts
and is used to define all the possible children (subloci, monoloci, loci, etc.)
"""

# Core imports
import collections
from operator import attrgetter

import networkx
from Mikado.serializers.blast_serializer import Hsp
from sqlalchemy import bindparam
from sqlalchemy.engine import Engine
from sqlalchemy.ext import baked
from sqlalchemy.orm.session import sessionmaker, Session
from sqlalchemy.sql.expression import and_
from .abstractlocus import Abstractlocus
from .monosublocusholder import MonosublocusHolder
from .sublocus import Sublocus
from ..exceptions import NotInLocusError
from ..parsers.GFF import GffLine
from ..serializers.blast_serializer import Hit, Query, Target
from ..serializers.external import External, ExternalSource
from ..serializers.junction import Junction
from ..serializers.orf import Orf
from ..transcripts import Transcript
from ..utilities import dbutils, grouper
from ..scales.assignment.assigner import Assigner
import bisect
from sys import maxsize
import functools
from collections import OrderedDict as SortedDict
from .locus import Locus
from .excluded import Excluded
from typing import Union, List, Dict
from ..utilities import Interval, IntervalTree
from itertools import combinations
import random
import asyncio

# The number of attributes is something I need
# pylint: disable=too-many-instance-attributes


bakery = baked.bakery()

junction_baked = bakery(lambda session: session.query(Junction))
junction_baked += lambda q: q.filter(and_(
    Junction.chrom == bindparam("chrom"),
    Junction.junction_start >= bindparam("junctionStart"),
    Junction.junction_end <= bindparam("junctionEnd")
))

hit_baked = bakery(lambda session: session.query(Hit))
hit_baked += lambda q: q.filter(and_(
    Hit.query_id.in_(bindparam("queries", expanding=True)),
    Hit.evalue <= bindparam("evalue"),
    Hit.hit_number <= bindparam("hit_number")
))

hsp_baked = bakery(lambda session: session.query(Hsp))
hsp_baked += lambda q: q.filter(and_(Hsp.hsp_evalue <= bindparam("hsp_evalue"),
                                     Hsp.query_id.in_(bindparam("queries", expanding=True))))


query_baked = bakery(lambda session: session.query(Query))
query_baked += lambda q: q.filter(Query.query_name.in_(bindparam("tids", expanding=True)))

target_baked = bakery(lambda session: session.query(Target))
target_baked += lambda q: q.filter(Target.target_id.in_(bindparam("targets", expanding=True)))

source_bakery = bakery(lambda session: session.query(ExternalSource))
source_bakery += lambda q: q.filter(ExternalSource.source.in_(bindparam("sources", expanding=True)))

orfs_baked = bakery(lambda session: session.query(Orf))
orfs_baked += lambda q: q.filter(Orf.query_id.in_(bindparam("queries", expanding=True)))


class Superlocus(Abstractlocus):
    """The superlocus class is used to define overlapping regions
    on the genome, and it receives as input transcript class instances.
    """

    __name__ = "superlocus"



    _complex_limit = (5000, 5000)

    # ###### Special methods ############

    def __init__(self,
                 transcript_instance,
                 verified_introns=None,
                 stranded=True,
                 configuration=None,
                 source="",
                 logger=None,
                 **kwargs):

        """

        :param transcript_instance: an instance of the Transcript class
        :type transcript_instance: [Transcript|None]
        :param stranded: boolean flag that indicates whether
        the Locus should use or ignore strand information
        :type stranded: bool
        :param configuration: a configuration dictionary derived from JSON/YAML config files
        :type configuration: (MikadoConfiguration|DaijinConfiguration)
        :param source: optional source for the locus
        :type source: str
        :param logger: the logger for the class
        :type logger: logging.Logger

        The superlocus class is instantiated from a transcript_instance class,
        which it copies in its entirety.

        It will therefore have the following attributes:
        - chrom, strand, start, end
        - splices - a *set* which contains the position of each splice site
        - introns - a *set* which contains the positions of each
        *splice junction* (registered as 2-tuples)
        - transcripts - a *set* which holds the transcripts added to the superlocus

        The constructor method takes the following keyword arguments:
        - stranded    True if all transcripts inside the superlocus are required
        to be on the same strand
        - configuration    Required. Configuration object for the run.
        - purge        Flag. If True, all loci holding only transcripts with a 0
        score will be deleted
        from further consideration.
        """

        super().__init__(source=source,
                         transcript_instance=None,
                         verified_introns=verified_introns,
                         configuration=configuration,
                         logger=logger,
                         **kwargs)
        self.approximation_level = 0
        self.feature = self.__name__
        self.stranded = stranded
        self.splices = set(self.splices)
        self.introns = set(self.introns)

        # Flags
        self.subloci_defined = False
        self.monosubloci_defined = False
        self.loci_defined = False
        self.monosubloci_metrics_calculated = False

        # Objects used during computation
        self.subloci = []
        self.loci = SortedDict()
        self.monosubloci = {}
        self.monoholders = []

        # Connection objects
        self.engine = self.sessionmaker = self.session = None
        # Excluded object
        self.excluded = Excluded(configuration=self.configuration)
        self.__data_loaded = False
        self.__lost = {}
        if transcript_instance is not None:
            super().add_transcript_to_locus(transcript_instance)
            assert transcript_instance.monoexonic is True or len(self.introns) > 0
            if self.stranded is True:
                self.strand = transcript_instance.strand
        else:
            self.logger.warning("Creating an empty superlocus as no transcript was provided!")

    def __create_locus_lines(self, superlocus_line: GffLine, new_id: str, print_cds=True):

        """
        Private method to prepare the lines for printing out loci
        into GFF/GTF files.
        """

        lines = []
        self.define_loci()
        if len(self.loci) > 0:
            source = "{0}_loci".format(self.source)
            superlocus_line.source = source
            lines.append(str(superlocus_line))
            found = dict()

            for _, locus_instance in self.loci.items():
                locus_instance.source = source
                locus_instance.parent = new_id
                if locus_instance.id in found:
                    found[locus_instance.id] += 1
                    locus_instance.counter = found[locus_instance.id]
                else:
                    found[locus_instance.id] = 0
                lines.append(locus_instance.__str__(print_cds=print_cds).rstrip())
        return lines

    def __create_monolocus_holder_lines(self, superlocus_line: GffLine, new_id: str, print_cds=True):

        """
        Private method to prepare the lines for printing out monosubloci
        into GFF/GTF files.
        """

        lines = []
        self.define_monosubloci()
        if len(self.monoholders) > 0:
            source = "{0}_monosubloci".format(self.source)
            superlocus_line.source = source
            lines.append(str(superlocus_line))
            found = dict()
            for monosublocus_instance in self.monoholders:
                monosublocus_instance.source = source
                monosublocus_instance.parent = new_id
                if monosublocus_instance.id in found:
                    found[monosublocus_instance.id] += 1
                    monosublocus_instance.counter = found[monosublocus_instance.id]
                else:
                    found[monosublocus_instance.id] = 0

                lines.append(monosublocus_instance.__str__(print_cds=print_cds).rstrip())

        return lines

    def __create_sublocus_lines(self, superlocus_line: GffLine, new_id: str, print_cds=True):
        """
        Private method to prepare the lines for printing out subloci
        into GFF/GTF files.
        """

        source = "{0}_subloci".format(self.source)
        superlocus_line.source = source
        lines = [str(superlocus_line)]
        self.define_subloci()
        found = dict()
        for sublocus_instance in self.subloci:
            assert hasattr(sublocus_instance, "source"), sublocus_instance
            sublocus_instance.source = source
            sublocus_instance.parent = new_id
            if sublocus_instance.id in found:
                found[sublocus_instance.id] += 1
                sublocus_instance.counter = found[sublocus_instance.id]
            else:
                found[sublocus_instance.id] = 0
            lines.append(sublocus_instance.__str__(print_cds=print_cds).rstrip())
        return lines

    # This discrepancy with the base class is necessary
    # pylint: disable=arguments-differ

    def format(self, print_cds=True, level=None):

        """
        Alias for __str__.
        :param print_cds: Boolean. It indicates whether to print the CDS features or not.
        :param level: level which we wish to print for. Can be "loci", "subloci", "monosubloci"
        :return: formatted GFF strings
        """
        return self.__str__(print_cds=print_cds, level=level)

    def __str__(self, level=None, print_cds=True):

        """
        This function will return the desired level of children loci.
        The keyword level accepts the following four values:
        - "None" - print whatever is available.
        - "loci" - print the final loci
        - "monosubloci" - print the monosubloci
        - "subloci" - print the subloci.

        The function will then return the desired location in GFF3-compliant format.

        :param level: level which we wish to print for. Can be "loci", "subloci", "monosubloci"
        :type level: str
        :param print_cds: flag. If set to False, only the exonic information will be printed.
        :type print_cds: bool
        """

        if abs(self.start) == float("inf") or abs(self.start) == maxsize:
            return ''

        assert level in (None, "loci", "subloci", "monosubloci"), f"Unrecognized level: {level}"

        superlocus_line = GffLine('')
        superlocus_line.chrom = self.chrom
        superlocus_line.feature = self.__name__
        superlocus_line.start, \
            superlocus_line.end, \
            superlocus_line.score = self.start, self.end, "."
        superlocus_line.strand = self.strand
        superlocus_line.phase, superlocus_line.score = None, None
        new_id = "{0}_{1}".format(self.source, self.id)
        superlocus_line.id, superlocus_line.name = new_id, self.name
        if self.approximation_level > 0:
            superlocus_line.attributes["approximation_level"] = self.approximation_level

        lines = []

        if level == "loci" or (level is None and self.loci_defined is True):
            lines = self.__create_locus_lines(
                superlocus_line,
                new_id,
                print_cds=print_cds
            )
        elif level == "monosubloci" or (level is None and self.monosubloci_defined is True):
            lines = self.__create_monolocus_holder_lines(superlocus_line,
                                                         new_id,
                                                         print_cds=print_cds)
        elif level == "subloci" or (level is None and self.monosubloci_defined is False):
            lines = self.__create_sublocus_lines(superlocus_line,
                                                 new_id,
                                                 print_cds=print_cds)
        if len(lines) > 0:
            lines.append("###")
        return "\n".join([line for line in lines if line is not None and line != ''])
    # pylint: enable=arguments-differ

    def add_locus(self, locus: Locus):
        """Method to add a Locus instance to the Superlocus instance."""

        self.add_transcripts_to_locus(locus.transcripts.values(), unsafe=True, check_in_locus=False)
        self.configuration = locus.configuration
        self.loci[locus.id] = locus
        self.loci_defined = True

    def as_dict(self, with_subloci=True, with_monoholders=True):
        """
        Method to dump a superlocus into a dictionary representation.
        :param with_subloci: boolean. Should we include the subloci in the dump?
        :param with_monoholders: boolean. Should we include the monoholders in the dump?
        :return:
        """

        state = super().as_dict()
        state["start"], state["end"] = self.start, self.end
        if with_subloci is True:
            state["subloci"] = [sublocus.as_dict() for sublocus in self.subloci]
        else:
            state["subloci"] = []
        state["loci"] = dict((lid, locus.as_dict()) for lid, locus in self.loci.items())
        if with_monoholders is True:
            state["monoholders"] = [mono.as_dict() for mono in self.monoholders]
        else:
            state["monoholders"] = []
        state["excluded"] = self.excluded.as_dict()
        return state

    def load_dict(self, state, print_subloci=True, print_monoloci=True, load_transcripts=True,
                  load_configuration=True):
        """Method to reconstitute a Superlocus from a dumped dictionary."""

        super().load_dict(state, load_transcripts=load_transcripts, load_configuration=load_configuration)
        self.loci = dict()
        for lid, stat in state["loci"].items():
            locus = Locus()
            locus.load_dict(stat)
            self.loci[lid] = locus

        if print_subloci is True:
            self.excluded = Excluded(configuration=self.configuration)
            self.excluded.load_dict(state["excluded"])
            self.subloci = []
            for stat in state["subloci"]:
                sub = Sublocus(configuration=self.configuration)
                sub.load_dict(stat)
                assert isinstance(sub, Sublocus)
                self.subloci.append(sub)
        else:
            self.subloci = []
            self.excluded = Excluded(configuration=self.configuration)

        if print_monoloci is True:
            self.monoholders = []
            for stat in state["monoholders"]:
                sub = MonosublocusHolder()
                sub.load_dict(stat)
                self.monoholders.append(sub)
        else:
            self.monoholders = []
        self.chrom, self.strand, self.start, self.end = state["chrom"], state["strand"], state["start"], state["end"]
        if len(self.loci) > 0 or len(self.transcripts) > 0:
            assert self.start != maxsize
            assert not self.id.endswith(str(maxsize))

    # ########### Class instance methods ############

    def _divide_by_coordinate(self, cdnas):
        if len(cdnas) == 0:
            return []
        cdnas = sorted(cdnas)
        cdnas[0].finalize()
        groups = [[cdnas[0]]]
        start, end = cdnas[0].start, cdnas[0].end
        for cdna in cdnas[1:]:
            cdna.finalize()
            assert cdna.chrom == self.chrom, (cdna.id, cdna.chrom)
            if self.overlap((cdna.start, cdna.end), (start, end),
                            flank=self.configuration.pick.clustering.flank) > 0:
                groups[-1].append(cdna)
                start, end = min(start, cdna.start), max(end, cdna.end)
            else:
                groups.append([cdna])
                start, end = cdna.start, cdna.end
        return groups

    def split_strands(self):
        """This method will divide the superlocus on the basis of the strand.
        The rationale is to parse a GFF file without regard for the
        strand, in order to find all intersecting loci;
        and subsequently break the superlocus into the different components.
        Notice that each strand might generate more than one superlocus,
        if genes on a different strand link what are
        two different superloci.
        """

        self.logger.debug("Splitting by strand for {0}".format(self.id))
        if self.stranded is True:
            self.logger.warning("Trying to split by strand a stranded Locus, {0}!".format(self.id))
            yield self
            return

        strands = {'+': [], '-': [], None: []}
        for cdna in self.transcripts.values():
            strands[cdna.strand].append(cdna)
        
        new_loci = []
        assert len(strands) == 3
        assert sum(len(_) for _ in strands.values()) == len(self.transcripts)
        for strand, cdnas in strands.items():
            for group in self._divide_by_coordinate(cdnas):
                new_locus = Superlocus(group[0],
                                       stranded=True,
                                       configuration=self.configuration,
                                       source=self.source,
                                       flank=self.configuration.pick.clustering.flank,
                                       logger=self.logger)
                new_locus.add_transcripts_to_locus(group[1:],
                                                   unsafe=True,
                                                   check_in_locus=True,
                                                   overwrite=False)
                assert len(new_locus.introns) > 0 or new_locus.monoexonic is True
                assert set(new_locus.transcripts.keys()) == {cdna.id for cdna in group}
                new_loci.append(new_locus)

        self.logger.debug(f"Defined {len(new_loci)} superloci by splitting {self.id} by strand.")
        yield from iter(sorted(new_loci))

    def connect_to_db(self, engine: Engine, session: Session):

        """
        :param engine: the connection pool
        :type engine: Engine

        :param session: a preformed session
        :type session: Session

        This method will connect to the database using the information
        contained in the JSON configuration.
        """

        if isinstance(session, Session):
            self.session = session
            self.sessionmaker = sessionmaker()
            self.sessionmaker.configure(bind=self.session.bind)
            self.engine = self.session.bind

        if engine is None:
            self.engine = dbutils.connect(self.configuration)
        else:
            self.engine = engine

        self.sessionmaker = sessionmaker()
        self.sessionmaker.configure(bind=self.engine)
        self.session = self.sessionmaker()

    def load_transcript_data(self, tid, data_dict):
        """
        :param tid: the name of the transcript to retrieve data for.
        :type tid: str

        :param data_dict: the dictionary to use for data retrieval, if specified.
        If None, a DB connection will be established to retrieve the necessary data.
        :type data_dict: (None | dict)

        This routine is used to load data for a single transcript."""

        self.logger.debug("Retrieving data for {0}".format(tid))
        self.transcripts[tid].logger = self.logger
        self.transcripts[tid].default_configuration = self.configuration
        self.transcripts[tid].load_information_from_db(self.configuration,
                                                       introns=self.locus_verified_introns,
                                                       data_dict=data_dict)
        to_remove, to_add = False, dict()

        if self.configuration.pick.chimera_split.execute is True:
            if self.transcripts[tid].number_internal_orfs > 1:
                new_tr = list(self.transcripts[tid].split_by_cds())
                if len(new_tr) > 1:
                    for new in new_tr:
                        assert new.id not in to_add
                        to_add[new.id] = new
                    to_remove = True
                    self.logger.info("%s has been split into %d different transcripts.",
                                     tid, len(new_tr))

        del data_dict
        return to_remove, to_add

    async def _load_introns(self):

        """Private method to load the intron data into the locus.
        :param data_dict: Dictionary containing the preloaded data, if available.
        :return:
        """

        if len(self.introns) == 0:
            assert self.monoexonic is True, f"{self.id} is multiexonic but has no introns defined!"
            self.logger.debug("No introns for %s", self.id)
            return

        self.logger.debug("Querying the DB for introns, %d total", len(self.introns))
        if not self.configuration.db_settings.db:
            return  # No data to load

        ver_introns = collections.defaultdict(set)
        for junc in junction_baked(self.session).params(chrom=self.chrom,
                                                        junctionStart=self.start, junctionEnd=self.end):
            ver_introns[(junc.junction_start, junc.junction_end)].add(junc.strand)

        self.logger.debug("Found %d verifiable introns for %s",
                          len(ver_introns), self.id)

        for intron in self.introns:
            self.logger.debug("Checking %s%s:%d-%d",
                              self.chrom, self.strand, intron[0], intron[1])
            if (intron[0], intron[1]) in ver_introns:
                if self.stranded is False:
                    for strand in ver_introns[(intron[0], intron[1])]:
                        self.locus_verified_introns.add((intron[0],
                                                         intron[1],
                                                         strand))
                elif self.strand in ver_introns[(intron[0], intron[1])]:
                    self.locus_verified_introns.add((intron[0],
                                                     intron[1],
                                                     self.strand))

    async def get_sources(self):
        if self.configuration.pick.output_format.report_all_external_metrics is True:
            sources = dict((source.source_id, source) for source in self.session.query(ExternalSource))
        else:
            sources = set()
            sources.update({param for param in self.configuration.scoring.requirements.parameters.keys()
                            if param.startswith("external")})
            sources.update({param for param in self.configuration.scoring.not_fragmentary.parameters.keys()
                            if param.startswith("external")})
            sources.update({param for param in self.configuration.scoring.cds_requirements.parameters.keys()
                            if param.startswith("external")})
            sources.update({param for param in self.configuration.scoring.as_requirements.parameters.keys()
                            if param.startswith("external")})
            sources.update({param for param in self.configuration.scoring.scoring.keys()
                            if param.startswith("external")})
            sources = {param.replace("external.", "") for param in sources}
            sources = dict((source.source_id, source) for source in
                           source_bakery(self.session).params(sources=list(sources)))
        return sources

    async def get_external(self, query_ids, qids):
        external = collections.defaultdict(dict)
        sources = await self.get_sources()
        baked = External.__table__.select().where(and_(External.source_id.in_(list(sources.keys())),
                                                       External.query_id.in_(qids)))
        for ext in self.session.execute(baked):
            source_id, query_id, score = ext.source_id, ext.query_id, ext.score
            assert source_id in sources and query_id in qids
            rtype = sources[source_id].rtype
            assert rtype in ("int", "float", "bool"), f"Invalid rtype: {rtype}"
            if rtype == "int":
                score = int(score)
            elif rtype == "float":
                score = float(score)
            elif rtype == "bool":
                score = bool(int(score))
            external[query_ids[ext.query_id].query_name][
                sources[ext.source_id].source] = (score, sources[ext.source_id].valid_raw)
        return external

    async def get_hits(self, query_ids, qids):
        hsps = dict()
        targets = set()
        hits = collections.defaultdict(list)
        for hsp in hsp_baked(self.session).params(
                hsp_evalue=self.configuration.pick.chimera_split.blast_params.hsp_evalue,
                queries=qids):
            if hsp.query_id not in hsps:
                hsps[hsp.query_id] = collections.defaultdict(list)
            hsps[hsp.query_id][hsp.target_id].append(hsp)
            targets.add(hsp.target_id)

        target_ids = dict((target.target_id, target) for target in
                          target_baked(self.session).params(targets=list(targets)))
        current_hit = None
        for hit in hit_baked(self.session).params(
                evalue=self.configuration.pick.chimera_split.blast_params.evalue,
                hit_number=self.configuration.pick.chimera_split.blast_params.max_target_seqs,
                queries=qids):
            if current_hit != hit.query_id:
                current_hit = hit.query_id
            current_counter = 0

            current_counter += 1

            my_query = query_ids[hit.query_id]
            my_target = target_ids[hit.target_id]

            hits[my_query.query_name].append(
                Hit.as_full_dict_static(
                    hit,
                    hsps[hit.query_id][hit.target_id],
                    my_query,
                    my_target
                )
            )
        return hits

    async def get_orfs(self, qids) -> Dict[str, List]:
        orfs = collections.defaultdict(list)
        for orf in orfs_baked(self.session).params(queries=qids):
            orfs[orf.query].append(orf.as_bed12())
        return orfs

    async def _create_data_dict(self, engine, tid_keys) -> dict:

        """Private method to retrieve data from the database and prepare it to be passed to the transcript
        instances.

        :param engine: the sqlalchemy engine to use
        :type engine: sqlalchemy.engine.Engine

        :param tid_keys: the transcript IDs inside the locus
        :type tid_keys: (set|list|tuple)

        """

        data_dict = dict()
        self.logger.debug("Starting to load hits and orfs for %d transcripts",
                          len(tid_keys))

        data_dict["external"] = collections.defaultdict(dict)
        data_dict["orfs"] = collections.defaultdict(list)
        data_dict["hits"] = collections.defaultdict(list)
        if engine is None:
            return data_dict

        for tid_group in grouper(tid_keys, 100):
            query_ids = dict((query.query_id, query) for query in
                             self.session.query(Query).filter(
                                 Query.query_name.in_(tid_group)))
            qids = list(query_ids.keys())
            data_dict["orfs"].update(await self.get_orfs(qids))
            data_dict["external"].update(await self.get_external(query_ids, qids))
            data_dict["hits"].update(await self.get_hits(query_ids=query_ids, qids=qids))
        return data_dict

    def load_all_transcript_data(self, engine=None, session=None):

        """
        This method will load data into the transcripts instances,
        and perform the split_by_cds if required
        by the configuration.
        Asyncio coroutines are used to decrease runtime.

        :param engine: a connection engine
        :type engine: Engine
        """

        if self.__data_loaded is True:
            return

        # Before starting to load: let's reduce the loci.
        if len(self.transcripts) >= self._complex_limit[0]:
            self.logger.warning("Complex locus %s with %d transcripts", self.id, len(self.transcripts))
            self.approximation_level = 1
            self.reduce_method_one(None)

        self.connect_to_db(engine, session)

        tid_keys = list(self.transcripts.keys())
        to_remove, to_add = set(), dict()
        # This will function even if data_dict is None
        intron_loader = self._load_introns()
        data_dict = self._create_data_dict(engine, tid_keys)
        data_dict = asyncio.run(data_dict)
        asyncio.run(intron_loader)

        self.logger.debug("Verified %d introns for %s",
                          len(self.locus_verified_introns),
                          self.id)

        self.logger.debug("Finished retrieving data for %d transcripts",
                          len(tid_keys))
        # self.session.close()
        # sasession.close_all_sessions()

        for tid in tid_keys:
            remove_flag, new_transcripts = self.load_transcript_data(tid, data_dict)
            if remove_flag is True:
                to_remove.add(tid)
                to_add.update(new_transcripts)

        assert len(to_remove) < len(to_add) + len(self.transcripts)

        if to_remove:
            self.logger.debug("Rebuilding the superlocus as %d transcripts have been excluded or split.",
                              len(to_remove))
            self.remove_transcripts_from_locus(to_remove)
            self.logger.debug(f"Adding {', '.join(to_add.keys())} to self.id")
            self.add_transcripts_to_locus(to_add.values(),
                                          unsafe=True,
                                          check_in_locus=False)
            assert set.issubset(set(to_add.keys()), set(self.transcripts.keys())), (
                f"Failed to add {len(to_add)} transcripts to {self.id}!"
            )

        elif len(to_remove) == len(self.transcripts):
            self.logger.warning("No transcripts left for %s", self.name)
            self.excluded.add_transcripts_to_locus(to_remove, unsafe=True)
            self._remove_all()

        del data_dict

        num_coding = 0
        for tid in self.transcripts:
            if self.transcripts[tid].combined_cds_length > 0:
                num_coding += 1
            else:
                self.transcripts[tid].feature = "ncRNA"

        self.logger.debug(
            "Found %d coding transcripts out of %d in %s",
            num_coding,
            len(self.transcripts),
            self.id)

        self.session = None
        self.__data_loaded = True
        self.sessionmaker = None
        self.stranded = False

    # ##### Sublocus-related steps ######

    def reduce_complex_loci(self, transcript_graph: networkx.Graph):

        """
        Method which checks whether a locus has too many transcripts and tries to reduce them.

        :param transcript_graph: the transcript graph to analyse for redundancies
        :return:
        """

        max_edges = max([d for n, d in transcript_graph.degree])
        self.approximation_level = 0
        if len(transcript_graph) < self._complex_limit[0] and max_edges < self._complex_limit[1]:
            return transcript_graph
        self.logger.warning("Complex superlocus with %d nodes \
        with the most connected having %d edges",
                            len(transcript_graph), max_edges)

        self.approximation_level = 1
        transcript_graph, max_edges = self.reduce_method_one(transcript_graph)

        if len(transcript_graph) < self._complex_limit[0] and max_edges < self._complex_limit[1]:
            self.logger.warning("Approximation level 1 for %s", self.id)
            self.logger.debug("Remaining transcripts: %s", ", ".join(self.transcripts.keys()))
            return transcript_graph

        self.logger.warning(
            "Still %d nodes with the most connected with %d edges after approximation 1",
            len(transcript_graph), max_edges)

        self.approximation_level = 2
        transcript_graph, max_edges = self.reduce_method_two(transcript_graph)
        # if len(transcript_graph) < self._complex_limit[0] and max_edges < self._complex_limit[1]:
        self.logger.warning("Approximation level 2 for %s", self.id)
        return transcript_graph

    def reduce_method_one(self, transcript_graph: Union[None, networkx.Graph]) -> [networkx.Graph, int]:

        """Approximation level one: we are going to group together all transcripts that have identical intron chains,
        and remove any that is completely contained within the longest ones. Reference transcripts get an automatic
        pass."""

        ichains = collections.defaultdict(list)
        cds_only = self.configuration.pick.clustering.cds_only
        for transcript in self.transcripts.values():
            # Ordered by coordinates
            if cds_only:
                bisect.insort(ichains[tuple(sorted(transcript.selected_cds_introns))], transcript)
            else:
                bisect.insort(ichains[tuple(sorted(transcript.introns))], transcript)

        to_remove = set()
        for ichain, transcripts in ichains.items():
            current_coords = [transcripts[0].start, transcripts[0].end]
            current_id = transcripts[0].id
            for transcript in transcripts[1:]:  # we know that they are sorted left to right
                self.logger.debug("Comparing %s to %s", current_id, transcript.id)
                if transcript_graph and transcript.id not in transcript_graph.neighbors(current_id):
                    self.logger.debug("%s and %s are not neighbours.", current_id, transcript.id)
                    continue
                if transcript.end <= current_coords[1] and transcript.start > current_coords[0]:
                    if transcript.is_reference:
                        continue
                    else:
                        self.logger.debug("Removing %s as it is contained", transcript.id)
                        to_remove.add(transcript.id)
                elif transcript.end == current_coords[1] and transcript.start == current_coords[0]:
                    if transcript.is_reference:
                        current_id = transcript.id
                        if not self.transcripts[current_id].is_reference:
                            self.logger.debug("Removing %s as it is identical to a reference", current_id)
                            to_remove.add(current_id)
                    elif self.transcripts[current_id].is_reference:
                        self.logger.debug("Removing %s as it is identical to a reference", transcript.id)
                        to_remove.add(transcript.id)
                    else:
                        # chosen = np.random.choice(sorted([current_id, transcript.id]))
                        chosen = random.choice(sorted([current_id, transcript.id]))
                        if current_id == chosen:
                            to_remove.add(transcript.id)
                        else:
                            to_remove.add(chosen)
                            current_id = transcript.id
                elif transcript.end > current_coords[1]:
                    self.logger.debug("Removing %s as it is contained in %s", current_id, transcript.id)
                    if transcript.start == current_coords[0] and not self.transcripts[current_id].is_reference:
                        to_remove.add(current_id)
                    current_coords, current_id = (transcript.start, transcript.end), transcript.id

        if to_remove:
            if transcript_graph:
                transcript_graph.remove_nodes_from(to_remove)
            self.excluded.add_transcripts_to_locus([self.transcripts[tid] for tid in to_remove],
                                                   check_in_locus=False, unsafe=True)
            self.logger.debug("Removing the following transcripts from %s: %s",
                              self.id, ", ".join(to_remove))

            self.remove_transcripts_from_locus(to_remove)

        max_edges = max([d for n, d in transcript_graph.degree]) if transcript_graph else 0
        return transcript_graph, max_edges

    def reduce_method_two(self, transcript_graph: networkx.Graph) -> [networkx.Graph, int]:

        """Approximation level two: we are going to remove transcripts that are contained within others. So e.g.
        a transcript with two exons might be seen as redundant with, and therefore removed, with a transcript with
        three or more exons."""

        to_remove = set()
        done = set()
        cds_only = self.configuration.pick.clustering.cds_only
        if cds_only:
            order = sorted(list(transcript_graph.nodes), key=lambda node: self.transcripts[node].selected_cds_start)
        else:
            order = sorted(list(transcript_graph.nodes), key=lambda node: self.transcripts[node].start)

        for tid in order:
            current = self.transcripts[tid]
            for neighbour in transcript_graph.neighbors(tid):
                couple = tuple(sorted((neighbour, tid)))
                if couple in done:
                    continue
                done.add(couple)
                self.logger.debug("Comparing %s to %s", tid, neighbour)
                if neighbour in to_remove:
                    continue
                neighbour = self.transcripts[neighbour]
                inters = set.intersection(current.introns, neighbour.introns)
                if inters == current.introns:
                    self.logger.debug("Evaluating %s (template %s) for removal", current.id, neighbour.id)
                    if current.is_reference:
                        self.logger.debug("Ignoring %s for removal as it is a reference transcript", current.id)
                        continue
                    if cds_only:
                        comparison = Assigner.compare(current.copy().remove_utrs(),
                                                      neighbour.copy().remove_utrs())[0]
                    else:
                        comparison = Assigner.compare(current, neighbour)[0]
                    self.logger.debug(
                        "Evaluating %s (template %s) for removal (%s)",
                        current.id, neighbour.id, (comparison.n_prec[0], comparison.n_recall[0]))
                    if comparison.n_prec[0] == 100:
                        self.logger.debug("Removing %s as it is contained in %s", current.id, neighbour.id)
                        to_remove.add(current.id)
                        break
                    self.logger.debug("%s is not perfectly contained in %s", current.id, neighbour.id)
                if inters == neighbour.introns:
                    if neighbour.is_reference:
                        self.logger.debug("Ignoring %s for removal as it is a reference transcript", neighbour.id)
                        continue
                    if cds_only:
                        comparison = Assigner.compare(neighbour.copy().remove_utrs(),
                                                      current.copy().remove_utrs())[0]
                    else:
                        comparison = Assigner.compare(neighbour, current)[0]
                    self.logger.debug(
                        "Evaluating %s (template %s) for removal (%s)",
                        current.id, neighbour.id, (comparison.n_prec[0], comparison.n_recall[0]))
                    if comparison.n_prec[0] == 100:
                        self.logger.debug("Removing %s as it is contained in %s", neighbour.id, current.id)
                        to_remove.add(neighbour.id)
                        break
                    self.logger.debug("%s is not perfectly contained in %s", neighbour.id, current.id)
                else:
                    self.logger.debug("%s and %s have differing introns, ignoring the comparison",
                                      neighbour.id, current.id)
                    continue

        if to_remove:
            transcript_graph.remove_nodes_from(to_remove)
            self.excluded.add_transcripts_to_locus([self.transcripts[tid] for tid in to_remove],
                                                   unsafe=True)
            self.logger.debug("Removing the following transcripts from %s: %s", self.id, ", ".join(to_remove))
            self.remove_transcripts_from_locus(to_remove)

        max_edges = max(d for n, d in transcript_graph.degree)
        return transcript_graph, max_edges

    def define_subloci(self, check_requirements=True):
        """This method will define all subloci inside the superlocus.
        Steps:
            - Call the BronKerbosch algorithm to define cliques
            - Call the "merge_cliques" algorithm the merge the cliques.
            - Create "sublocus" objects from the merged cliques
            and store them inside the instance store "subloci"
        """

        if self.subloci_defined is True:
            return

        self.subloci = []

        # First, check whether we need to remove CDS from anything.
        for tid in super()._check_not_passing(section_name="cds_requirements"):
            self.transcripts[tid].strip_cds(strand_specific=True)
            self.metrics_calculated = False

        # Check whether there is something to remove
        if check_requirements is True:
            self._check_requirements()
            excluded_tids = list(self._excluded_transcripts.keys())
            self.excluded.add_transcripts_to_locus([self._excluded_transcripts[excluded_tid]
                                                    for excluded_tid in excluded_tids],
                                                   check_in_locus=False,
                                                   unsafe=True)
            to_remove = set.intersection(set(self.transcripts.keys()), set(excluded_tids))
            self.remove_transcripts_from_locus(to_remove)
            [self._excluded_transcripts.pop(excluded_tid, None) for excluded_tid in excluded_tids]

        if len(self.transcripts) == 0:
            # we have removed all transcripts from the Locus. Set the flag to True and exit.
            self.subloci_defined = True
            return

        self.logger.debug("Calculating the transcript graph for %d transcripts", len(self.transcripts))
        transcript_graph = self.define_graph()

        transcript_graph = self.reduce_complex_loci(transcript_graph)
        if len(self.transcripts) > len(transcript_graph) and self.reference_update is False:
            self.logger.warning("Discarded %d transcripts from %s due to approximation level %d",
                                len(self.transcripts) - len(transcript_graph),
                                self.id,
                                self.approximation_level)
            for tid in set.difference(set(self.transcripts.keys()), set(transcript_graph.nodes())):
                self.logger.debug("Discarding %s from %s", tid, self.id)
                del self.transcripts[tid]

        if len(self.transcripts) == 0:
            # we have removed all transcripts from the Locus. Set the flag to True and exit.
            self.logger.warning("Discarded all transcripts from %s", self.id)
            self.subloci_defined = True
            return

        self.logger.debug("Calculated the transcript graph for %d transcripts: %s",
                          len(self.transcripts),
                          str(transcript_graph))
        self.logger.debug("Calculating the transcript communities")
        subloci = self.find_communities(transcript_graph)
        self.logger.debug("Calculated the transcript communities")

        # Now we should define each sublocus and store it in a permanent structure of the class
        for subl in subloci:
            if len(subl) == 0:
                continue
            subl = [self.transcripts[x] for x in subl]
            subl = sorted(subl)
            assert self.configuration.scoring is not None
            assert hasattr(self.configuration.scoring.requirements, "parameters")
            self.logger.debug("Scoring file: %s", self.configuration.pick.scoring_file)
            self.logger.debug(self.configuration.scoring.requirements._expression)
            new_sublocus = Sublocus(subl[0],
                                    configuration=self.configuration,
                                    logger=self.logger,
                                    use_transcript_scores=self._use_transcript_scores
                                    )
            new_sublocus.logger = self.logger
            new_sublocus.add_transcripts_to_locus(subl[1:], unsafe=True)
            new_sublocus.parent = self.id
            new_sublocus.metrics_calculated = False
            new_sublocus.get_metrics()
            self.subloci.append(new_sublocus)

        self.subloci = sorted(self.subloci)
        self.subloci_defined = True

    def define_monosubloci(self, check_requirements=True):

        """This is a wrapper method that defines the monosubloci for each sublocus.
        """
        if self.monosubloci_defined is True:
            return

        self.logger.debug("Calculating subloci for %s, %d transcripts",
                          self.id, len(self.transcripts))
        self.define_subloci()
        self.logger.debug("Calculated subloci for %s, %d transcripts",
                          self.id, len(self.transcripts))
        self.monosubloci = {}
        # Extract the relevant transcripts
        for sublocus_instance in sorted(self.subloci):
            sublocus_instance.logger = self.logger
            sublocus_instance.define_monosubloci(purge=self.purge, check_requirements=check_requirements)
            self.excluded.add_transcripts_to_locus(sublocus_instance.excluded.transcripts.values(),
                                                   unsafe=True)
            for tid in sublocus_instance.transcripts:
                # Update the score
                self.transcripts[tid].score = sublocus_instance.transcripts[tid].score
            for monosubl in sublocus_instance.monosubloci:
                monosubl.parent = self.id
                self.monosubloci[monosubl.tid] = monosubl

        if self.logger.level == 10:  # DEBUG
            self.logger.debug(
                "Monosubloci for %s:\n\t\t%s",
                self.id,
                "\n\t\t".join(
                    "{}, transcript: {}".format(self.monosubloci[_].id, _)
                    for _ in self.monosubloci
                ),
            )

        self.monosubloci_defined = True

    def print_subloci_metrics(self):
        """Wrapper method to create a csv.DictWriter instance and call
        the sublocus.print_metrics method
        on it for each sublocus."""

        for slocus in self.subloci:
            yield from slocus.print_metrics()
        if self.excluded is not None:
            yield from self.excluded.print_metrics()

    def print_subloci_scores(self):
        """Wrapper method to create a csv.DictWriter instance and call the
        sublocus.print_metrics method
        on it for each sublocus."""

        for slocus in self.subloci:
            yield from slocus.print_scores()

    def print_monoholder_metrics(self):

        """Wrapper method to create a csv.DictWriter instance and call the
        MonosublocusHolder.print_metrics method
        on it."""

        self.define_monosubloci()
        for monoholder in self.monoholders:
            yield from monoholder.print_metrics()

    def print_monoholder_scores(self):

        """Wrapper method to create a csv.DictWriter instance and call
        the MonosublocusHolder.print_scores method on it."""

        self.define_monosubloci()
        for monoholder in self.monoholders:
            yield from monoholder.print_scores()

    def print_loci_metrics(self):

        self.define_loci()

        if len(self.loci) == 0:
            return []
        for locus in self.loci:
            yield from self.loci[locus].print_metrics()

    def print_loci_scores(self):

        """Wrapper method to create a csv.DictWriter instance and call
        the Locus.print_scores method on it."""

        self.define_loci()

        # self.available_monolocus_metrics = set(self.monoholder.available_metrics)
        if len(self.loci) == 0:
            return
        for locus in self.loci:
            yield from self.loci[locus].print_scores()

    def define_loci(self, check_requirements=True):
        """This is the final method in the pipeline. It creates a container
        for all the monosubloci (an instance of the class MonosublocusHolder)
        and retrieves the loci it calculates internally."""

        if self.loci_defined is True:
            return

        self.logger.debug("Calculating monosubloci for %s, %d transcripts",
                          self.id, len(self.transcripts))
        self.define_monosubloci(check_requirements=check_requirements)
        self.logger.debug("Calculated monosubloci for %s, %d transcripts",
                          self.id, len(self.transcripts))
        self.calculate_mono_metrics(check_requirements=check_requirements)

        self.loci = SortedDict()
        if len(self.monoholders) == 0:
            if len(self.transcripts) == 0:
                # We have already purged
                self.logger.debug("No locus retained for %s", self.id)
            else:
                self.logger.warning("No locus retained for %s", self.id)
            self.loci_defined = True
            return

        loci = []
        for monoholder in self.monoholders:
            monoholder.define_loci(check_requirements=check_requirements)
            for locus_instance in monoholder.loci:
                monoholder.loci[locus_instance].parent = self.id
                loci.append(monoholder.loci[locus_instance])

        for locus in sorted(loci):
            self.loci[locus.id] = locus
            self.loci[locus.id].logger = self.logger

        self.loci_defined = True

        self.logger.debug("Looking for AS events in %s: %s",
                          self.id,
                          self.configuration.pick.alternative_splicing.report)
        if self.configuration.pick.alternative_splicing.report is True:
            self.define_alternative_splicing()

        self._find_lost_transcripts()
        while len(self.lost_transcripts):
            lost_transcripts = list(self.lost_transcripts.values())
            new_locus = Superlocus(lost_transcripts[0],
                                   configuration=self.configuration,
                                   use_transcript_scores=self._use_transcript_scores,
                                   stranded=self.stranded,
                                   verified_introns=self.locus_verified_introns,
                                   logger=self.logger,
                                   source=self.source
                                   )
            new_locus.add_transcripts_to_locus(lost_transcripts[1:],
                                               check_in_locus=False,
                                               unsafe=True)
            new_locus.define_loci(check_requirements=check_requirements)
            self.loci.update(new_locus.loci)
            self.__lost = new_locus.lost_transcripts

        if self.configuration.pick.run_options.only_reference_update is True:
            lids = list(self.loci.keys())[:]
            for lid in lids:
                if self.loci[lid].has_reference_transcript is False:
                    self.logger.debug("Removing %s (primary: %s) as it has no reference transcripts.",
                                      lid, self.loci[lid].primary_transcript_id)
                    del self.loci[lid]
            self.logger.debug("Remaining loci in %s: %s", self.id, ",".join(list(self.loci.keys())))

        return

    def _find_lost_transcripts(self):

        """Private method to identify, after defining loci, all transcripts that are not intersecting any
        of the resulting genes and that therefore could constitute "lost" genes.
        This could happen if e.g. a valid transcript is deselected in the first stage ("sublocus") as it intersects a
        longer transcript, which will itself be subsequently deselected in comparison with another shorter one. This
        would leave the first transcript stranded.
        WARNING: if a transcript's score has fallen at or below 0, and self.purge is False, transcripts will *not*
        be recovered by this method. Mikado will presume that excluding them was the correct thing to do and ignore
        them henceforth.
        """

        self.__lost = dict()
        cds_only = self.configuration.pick.clustering.cds_only
        simple_overlap = self.configuration.pick.clustering.simple_overlap_for_monoexonic
        cdna_overlap = self.configuration.pick.clustering.min_cdna_overlap
        cds_overlap = self.configuration.pick.clustering.min_cds_overlap

        loci_transcripts = set()
        for locus in self.loci.values():
            loci_transcripts.update(set([_ for _ in locus.transcripts.keys()]))

        not_loci_transcripts = set.difference({_ for _ in self.transcripts.keys()
                                               if _ not in self._excluded_transcripts},
                                              loci_transcripts)

        for locus in self.loci.values():
            to_remove = set()
            for tid in not_loci_transcripts:
                found = False
                for ltid in locus.transcripts:
                    found = found or MonosublocusHolder.is_intersecting(
                        self.transcripts[tid],
                        locus[ltid],
                        cds_only=cds_only,
                        logger=self.logger,
                        min_cdna_overlap=cdna_overlap,
                        min_cds_overlap=cds_overlap,
                        simple_overlap_for_monoexonic=simple_overlap
                    )
                    if found:
                        self.logger.debug("%s intersects %s in %s, not lost.",
                                          tid, ltid, locus.id)
                        break
                if found:
                    to_remove.add(tid)
            not_loci_transcripts = set.difference(not_loci_transcripts, to_remove)

        self.__lost.update({tid: self.transcripts[tid] for tid in not_loci_transcripts
                            if (self.purge is False or self.transcripts[tid].score > 0)})

        if len(self.__lost):
            self.logger.debug("Lost %s transcripts from %s; starting the recovery process: TIDs: %s",
                              len(self.lost_transcripts), self.id, ", ".join(self.__lost))

    def define_alternative_splicing(self):

        """
         This method will consider all possible candidates for alternative splicing
         for each of the final loci, after excluding transcripts which potentially map
         to more than one Locus (in order to remove chimeras).
         It will then call the add_transcript_to_locus method to try to add
         the transcript to the relevant Locus container.
        """

        # First off, define genes

        self.define_loci()

        candidates = collections.defaultdict(set)
        primary_transcripts = set(locus.primary_transcript_id for locus in self.loci.values())
        self.logger.debug("Primary transcripts: %s", primary_transcripts)

        cds_only = self.configuration.pick.clustering.cds_only
        cds_overlap = self.configuration.pick.alternative_splicing.min_cds_overlap
        cdna_overlap = self.configuration.pick.alternative_splicing.min_cdna_overlap

        self.logger.debug("Defining the transcript graph")
        t_graph = self.define_as_graph(inters=MonosublocusHolder.is_intersecting,
                                       cds_only=cds_only,
                                       min_cdna_overlap=cdna_overlap,
                                       min_cds_overlap=cds_overlap,
                                       simple_overlap_for_monoexonic=False)

        self.logger.debug("Defined the transcript graph")

        loci_cliques = dict()
        for lid, locus_instance in self.loci.items():
            assert locus_instance.primary_transcript_id in t_graph.nodes()
            neighbors = set(t_graph.neighbors(locus_instance.primary_transcript_id))
            loci_cliques[lid] = neighbors
            self.logger.debug("Neighbours for %s: %s", lid, neighbors)

        for tid in iter(tid for tid in self.transcripts if tid not in primary_transcripts):
            loci_in = list(llid for llid in loci_cliques if
                           tid in loci_cliques[llid])
            if len(loci_in) == 1:
                candidates[loci_in[0]].add(tid)
            elif len(loci_in) > 1:
                # These are transcripts that match more than one locus
                continue
            else:
                # These transcripts have been lost.
                self.__lost.update({tid: self[tid]})

        for lid in candidates:
            self.loci[lid].add_transcripts_to_locus(
                sorted([self.transcripts[tid] for tid in candidates[lid]],
                       key=attrgetter("score"), reverse=True)
            )

        # Now we have to recheck that no AS event is linking more than one locus.
        to_remove = collections.defaultdict(set)
        for lid in self.loci:
            for tid, transcript in [_ for _ in self.loci[lid].transcripts.items() if
                                    _[0] != self.loci[lid].primary_transcript_id]:
                for olid in [_ for _ in self.loci if _ != lid]:
                    is_compatible = MonosublocusHolder.in_locus(self.loci[olid],
                                                                transcript)
                    if is_compatible is True:
                        self.logger.debug("%s is compatible with more than one locus. Removing it.", tid)
                        to_remove[lid].add(tid)

        for lid in to_remove:
            self.loci[lid].remove_transcripts_from_locus(to_remove[lid])

        # This is *necessary* because the names of the loci *MIGHT HAVE CHANGED*!
        new = dict()
        for lid in self.loci:
            self.loci[lid].finalize_alternative_splicing()
            new[lid] = self.loci[lid]

        self.loci = new
        assert all(locus.metrics_calculated is True for locus in self.loci.values()), \
            "Loci not properly marked as defined after defining the alternative splicing."
        return

    def calculate_mono_metrics(self, check_requirements=True):
        """Wrapper to calculate the metrics for the monosubloci."""
        self.monoholders = []
        self.define_monosubloci(check_requirements=check_requirements)

        mono_graph = super().define_graph(
            self.monosubloci,
            inters=MonosublocusHolder.in_locus,
            logger=self.logger,
            cds_only=self.configuration.pick.clustering.cds_only,
            min_cdna_overlap=self.configuration.pick.clustering.min_cdna_overlap,
            min_cds_overlap=self.configuration.pick.clustering.min_cds_overlap,
            simple_overlap_for_monoexonic=self.configuration.pick.clustering.simple_overlap_for_monoexonic)

        assert len(mono_graph.nodes()) == len(self.monosubloci)

        communities = self.find_communities(mono_graph)

        for community in communities:
            community = set(community)
            monosub = self.monosubloci[community.pop()]
            holder = MonosublocusHolder(monosub,
                                        configuration=self.configuration,
                                        logger=self.logger,
                                        use_transcript_scores=self._use_transcript_scores)
            while community:
                holder.add_monosublocus(self.monosubloci[community.pop()],
                                        check_in_locus=False)
            self.monoholders.append(holder)

        for monoholder in self.monoholders:
            # monoholder.scores_calculated = False
            monoholder.filter_and_calculate_scores(check_requirements=check_requirements)

    # ############ Class methods ###########

    # The discrepancy is by design
    # pylint: disable=arguments-differ

    def define_graph(self) -> networkx.Graph:

        """Calculate the internal exon-intron graph for the object."""

        graph = networkx.Graph()

        # As we are using intern for transcripts, this should prevent
        # memory usage to increase too much
        keys = set.difference(set(self.transcripts.keys()), set(self._excluded_transcripts))

        graph.add_nodes_from(keys)

        monos = []
        intronic = collections.defaultdict(set)

        for tid in keys:
            transcript = self.transcripts[tid]
            if self._cds_only is True and transcript.is_coding:
                if transcript.selected_cds_introns:
                    for intron in transcript.selected_cds_introns:
                        intronic[intron].add(tid)
                else:
                    start, end = sorted([transcript.selected_cds_start, transcript.selected_cds_end])
                    interval = Interval(start, end, transcript.id)
                    monos.append(interval)
                    monos.append(interval)
            else:
                if transcript.introns:
                    for intron in transcript.introns:
                        intronic[intron].add(tid)
                else:
                    interval = Interval(transcript.start, transcript.end, transcript.id)
                    monos.append(interval)

        edges = set()
        for intron in intronic:
            edges.update(set(combinations(intronic[intron], 2)))

        # Now the monoexonic
        monoexonic = IntervalTree()
        [monoexonic.add(interval) for interval in monos]
        monos = sorted(monos)
        for mono in monos:
            edges.update(set((mono.value, omono.value) for omono in
                             (other for other in monoexonic.find(mono[0], mono[1], strict=False)
                              if mono.value != other.value)))
        graph.add_edges_from(edges)

        return graph

    def define_as_graph(self,
                        inters=MonosublocusHolder.is_intersecting,
                        cds_only=False,
                        min_cdna_overlap=0.2,
                        min_cds_overlap=0.2,
                        simple_overlap_for_monoexonic=True):

        """This method will try to build the AS graph using a O(nlogn) rather than O(n^2) algorithm.

        :param inters: the "is_intersecting" function to use.
        :param cds_only: whether to only consider the CDS to verify if two transcripts are intersecting
        :type cds_only: bool
        :param simple_overlap_for_monoexonic: boolean flag. If set to True (default) then a monoexonic transcript
        will be considered as intersecting the other transcript if even a single base is overlapping.
        :param min_cdna_overlap: minimum cDNA overlap for two transcripts
        :param min_cds_overlap: minimum CDS overlap (if *both* transcripts are coding).
        """

        method = functools.partial(inters, cds_only=cds_only, min_cdna_overlap=min_cdna_overlap,
                                   min_cds_overlap=min_cds_overlap,
                                   simple_overlap_for_monoexonic=simple_overlap_for_monoexonic,
                                   logger=self.logger)

        graph = networkx.Graph()
        graph.add_nodes_from(self.transcripts.keys())

        itree = IntervalTree()
        primaries = set()

        for lid in self.loci:
            transcript = self.loci[lid].primary_transcript
            primaries.add(transcript.id)
            if transcript.is_coding and cds_only:
                start, end = sorted([transcript.selected_cds_start, transcript.selected_cds_end])
            else:
                start, end = transcript.start, transcript.end
            itree.add(Interval(start, end, transcript.id))

        for tid in self.transcripts:
            if tid in primaries:
                continue
            transcript = self[tid]
            if transcript.is_coding and cds_only:
                start, end = sorted([transcript.selected_cds_start, transcript.selected_cds_end])
            else:
                start, end = transcript.start, transcript.end
            for found in itree.find(start, end, strict=False):
                locus_transcript = found.value
                if method(self[locus_transcript], transcript):
                    graph.add_edge(tid, locus_transcript)

        return graph

    @classmethod
    def is_intersecting(cls, transcript: Transcript, other: Transcript, cds_only=False) -> bool:
        """
        When comparing two transcripts, for the definition of subloci inside superloci we follow these rules:

        If both are multiexonic, the function verifies whether there is at least one intron in common.
        If both are monoexonic, the function verifies whether there is some overlap between them.
        If one is monoexonic and the other is not, the function will return False by definition.

        :rtype : bool
        :param transcript: a transcript for which we wish to verify
        whether it is intersecting with another transcript or not.
        :type transcript: Transcript
        :param other: the transcript which will be used for the comparison.
        :type other: Transcript

        :param cds_only: boolean flag. If enabled, only CDS exons/intron will be considered when deciding whether
        two transcripts are part of the same Locus or not.
        :type cds_only: bool
        """

        transcript.finalize()
        other.finalize()
        if transcript == other:
            return False  # We do not want intersection with oneself

        if transcript.monoexonic is False and other.monoexonic is False:
            if all([cds_only, transcript.is_coding, other.is_coding]):
                intersection = set.intersection(transcript.selected_cds_introns,
                                                other.selected_cds_introns)
            else:
                intersection = set.intersection(transcript.introns, other.introns)

            intersecting = (len(intersection) > 0)
        elif transcript.monoexonic is True and other.monoexonic is True:
            if all([cds_only, transcript.is_coding, other.is_coding]):
                intersecting = (cls.overlap(
                    (transcript.selected_cds_start, transcript.selected_cds_end),
                    (other.selected_cds_start, other.selected_cds_end), positive=False) > 0)
            else:
                intersecting = (cls.overlap(
                    (transcript.start, transcript.end),
                    (other.start, other.end), positive=False) > 0)
        else:
            intersecting = False

        return intersecting
    # pylint: enable=arguments-differ

    # ############## Properties ############
    @property
    def id(self) -> str:
        """
        This is a generic string generator for all inherited children.
        :rtype : str
        """
        if self.stranded is True:
            strand = self.strand
        else:
            strand = "mixed"
        return "{0}:{1}{2}:{3}-{4}".format(
            self.__name__,
            self.chrom,
            strand,
            self.start,
            self.end)

    @property
    def lost_transcripts(self):
        return self.__lost.copy()



#!/usr/bin/env python3
# coding: utf-8

"""
Script to extract features from a GTF with certain coordinates.
"""

import argparse
import sys
from ...utilities import to_region


__author__ = 'Luca Venturini'


def launch(args):

    """
    Simple launcher script.

    :param args: the argparse Namespace
    """
    from ...transcripts.transcript import Transcript
    from ...parsers.GTF import GTF

    if hasattr(args, "region") and args.region is not None:
        try:
            args.chrom, args.start, args.end = args.region
        except ValueError as exc:
            raise ValueError("{0} {1}".format(exc, args.region))

    if args.start >= args.end:
        raise ValueError("Start greater than end: {0}\t{1}".format(args.start, args.end))

    transcript = None
    with GTF(args.gtf) as gtf:
        for row in gtf:
            if row.chrom != args.chrom:
                continue
            else:
                if row.is_transcript is True:
                    if transcript is not None and \
                            transcript.start >= args.start and transcript.end <= args.end:
                        print(transcript.format("gtf"), file=args.out)
                        transcript = None
                    if args.assume_sorted is True and row.start > args.end:
                        break
                    transcript = Transcript(row)
                else:
                    transcript.add_exon(row)

    if transcript is not None and transcript.start >= args.start and transcript.end <= args.end:
        print(transcript.format("gtf"), file=args.out)


def awk_parser():
    """
    Parser for the command line
    :return:
    """

    parser = argparse.ArgumentParser(
        "Utility to extract features from a GTF within given coordinates.")
    region = parser.add_mutually_exclusive_group(required=True)
    region.add_argument("-r", "--region", type=to_region,
                        help="Region defined as a string like <chrom>:<start>..<end>")
    region.add_argument("--chrom", type=str)
    parser.add_argument("-as", "--assume-sorted", dest="assume_sorted",
                        default=False, action="store_true")
    parser.add_argument("--start", type=int, default=float("-inf"))
    parser.add_argument("--end", type=int, default=float("inf"))

    parser.add_argument("gtf", type=argparse.FileType())
    parser.add_argument("out", type=argparse.FileType("w"), nargs="?", default=sys.stdout)
    parser.set_defaults(func=launch)
    return parser



#!/usr/bin/env python3
# coding: utf-8

"""Quick script to automate the generation of metrics definition from the files."""

import sys
from ...transcripts import Transcript
import argparse
import tabulate
import textwrap
from itertools import zip_longest
from inspect import getdoc


__author__ = 'Luca Venturini'


def launch(args):

    """
    Main caller.

    :param args: the argparse Namespace.
    """

    metric_names = Transcript.get_available_metrics()
    print(file=args.out)
    metrics = ["tid", "parent", "score", "external_scores"]
    metrics.extend(
        sorted(metric for metric in metric_names
               if metric not in ("tid", "parent", "score")))

    if len(args.metric) > 0:
        if not all(metric in metrics for metric in args.metric):
            print("Invalid metrics selected: {}".format(
                ", ".join(sorted(metric for metric in args.metric if metric not in metrics))))
        metrics = args.metric
    elif len(args.category) > 0:
        metrics = [metric for metric in Transcript.get_available_metrics() if
                   getattr(getattr(Transcript, metric), "category", "Descriptive") in args.category]

    rows = []

    for metric in metrics:
        docstring = getdoc(getattr(Transcript, metric))
        category = getattr(getattr(Transcript, metric), "category", "Descriptive")
        usable_raw = getattr(getattr(Transcript, metric), "usable_raw", False)
        rtype = getattr(getattr(Transcript, metric), "rtype", "str")

        if metric == "external_scores":
            usable_raw = True
            rtype = "Namespace"
            category = "External"

        if docstring is None:
            docstring = ''
        else:
            docstring = textwrap.wrap(docstring, 57)

        met_rows = zip_longest([metric], docstring, [category], [rtype], [str(usable_raw)])
        rows.extend(met_rows)

    __table_format = tabulate._table_formats[args.format]

    # TableFormat(lineabove=Line("+", "-", "+", "+"),
    #             linebelowheader=Line("+", "=", "+", "+"),
    #             linebetweenrows=Line("+", "-", "+", "+"),
    #             linebelow=Line("+", "-", "+", "+"),
    #             headerrow=DataRow("|", "|", "|"),
    #             datarow=DataRow("|", "|", "|"),
    #             padding=1, with_header_hide=None),
    # Line = namedtuple("Line", ["begin", "hline", "sep", "end"])
    #
    # DataRow = namedtuple("DataRow", ["begin", "sep", "end"])

    if args.format not in ("grid", "fancy_grid"):

        print(tabulate.tabulate(rows,
                                headers=["Metric name", "Description", "Category", "Data type", "Usable raw"],
                                tablefmt=args.format))
    else:
        out_of_header = False
        separator = None

        for row in tabulate.tabulate(rows,
                                     headers=["Metric name", "Description", "Category", "Data type", "Usable raw"],
                                     tablefmt=args.format).split("\n"):
            if row[:2] == __table_format.lineabove.begin + __table_format.lineabove.hline:
                separator = row
                if not out_of_header:
                    print(row)
                continue
            if row[:2] == __table_format.linebelowheader.begin + __table_format.linebelowheader.hline:
                out_of_header = True
                print(row, file=args.out)
                continue
            elif out_of_header is False:
                print(row, file=args.out)
            elif row[:2] == __table_format.linebetweenrows[0] + __table_format.linebetweenrows[1]:
                continue
            elif row[0] == __table_format.datarow.begin:
                if row.strip().split(__table_format.datarow.sep)[1].strip() != "":
                    print(separator, file=args.out)
                print(row, file=args.out)
        print(separator, file=args.out)
        print(file=args.out)
    return


def metric_parser():

    """
    Command line parser for the utility.
    """

    parser = argparse.ArgumentParser("Script to generate the available metrics")
    parser.add_argument("-f", "--format",
                        help="Format of the table to be printed out.",
                        choices=tabulate.tabulate_formats, default="rst")
    parser.add_argument("-o", "--out",
                        help="Optional output file",
                        type=argparse.FileType("w"), default=sys.stdout)
    parser.add_argument("-c", "--category",
                        help="Available categories to select from.",
                        default=[], nargs="+",
                        choices=sorted(set(
                            [_ for _ in [getattr(getattr(Transcript, metric), "category", "Descriptive") for metric in
                             Transcript.get_available_metrics()] if _ is not None] + ["Descriptive"])))
    parser.add_argument("metric", nargs="*")
    parser.set_defaults(func=launch)
    return parser
